# -*- coding: utf-8 -*-
"""
Created on Thu Aug 10 16:25:13 2023

@author: smvanessen
"""

# Clear all variables
from IPython import get_ipython
get_ipython().magic('reset -sf')

# Standard toolboxes
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import os

# Color functions
from matplotlib import cm
from matplotlib.colors import Normalize

# Close all figures
plt.close('all')


#############
# START INPUT
#############

# Select GPR version 'GPR' for single-fidelity, or 'MF-GPR' for multi-fidelity.
GPRversions = ['GPR','MF-GPR']

# Select used initial sampling option. Note: you can only select options for 
# which you first generated data with the Adaptive Screening method. For a 
# description of options A, B and C, also see Adaptive Screening. 
# Options: 'A', 'B' or 'C'
sampopt = 'B'

# Selected acquisition function for some of the figures. Note: you can only 
# select functions for which you first generated data with the Adaptive Screening 
# method. Options: 'US', 'USMV', 'USPD' or 'IVR'.
acqsel = 'USMV'

# Folder where the input CSV files are (generated by the Adaptive Screening procedure) (in)
# Folder where we write the output figures (out)
in_files = 'out_files_waves'
out_figs = 'out_figs_paper'

# Kernel name single-fidelity HF-GPR. Note: you can only select kernels with 
# which you first generated data with the Adaptive Screening method. Preferred: 
# 'Mat32' kernel for all three.
HFkern = 'Mat32'
MFkern1 = 'Mat32'
MFkern2 = 'Mat32'

# Target exceedance duration in [s]. In the paper we used 1 hour.
dur = 3600

# Derive target probability of exceedance, using the average number of wave 
# encounters in target duration. These numbers are valid for our test case 
# study from the paper*. Only 'ne_exp' is used in the remainder, peak wave 
# period Tp is not.
Tp = 11
ne_exp = dur/Tp 

# Input: true HF MPM value from the MCS. Obtained using the script 'Generate_Lin_2nd_MCS_data.py'
MPM_true_MCS = 10.267260806161483

# Maximum number of seeds to include in plot (x-axis scaling)
seedmaxplot = 205

# Colors and linestyles for the plots. cols assigned to different acquisition 
# functions, cols2 to different indicator quality cases.
linestyles = ['-','--','-.',':']
cols = ['royalblue','indianred','mediumseagreen','orange'] # assigned to the acquisition functions
cols2 = ['deepskyblue','maroon','deeppink','goldenrod'] # assigned to the cases

# Plotswitch for the distributions of all iterations in one plot (1 = plot, 0 = do not plot)
plotalldistroneplot = 1

# Number of iterations to take the mean and limit for max. absolute error E
# (in this case in [m] wave height) in the stopping criterion.
N_mae = 20
# lim_MAE_conv = 0.03
lim_MAE_conv = 0.05

# Number of iterations over which to take the coefficient of variation C 
# and associated limit in the stopping criterion.
N_cov = 20
# lim_cov_conv = 0.00098
lim_cov_conv = 0.003

# Max. number of available iterations for acquisition function USMV
count_USMV = 200
   

#############
# END INPUT
#############

# Check wether output directories exist, if not make them
isExist1 = os.path.exists(out_figs)
if not isExist1:
   os.makedirs(out_figs)

# Target probability of exceedance.
pexp = 1/ne_exp

# Initialise
plt.figure(30000,figsize=(7,5))

# Loop over GPR and MF-GPR
for gg,GPRvers in enumerate(GPRversions):


    # Properties attributed to the GPR version and initial sample option
    if GPRvers == 'GPR':
        gprcase = 'GPR'
    elif GPRvers == 'MF-GPR':
        gprcase = 'MF-GPR'
            
    # Initial sample options (see Adaptive Screening)
    if sampopt == 'A':
        sampopt_txt = 'A_LFsubsamp'; N_hf = 2
        txtadd = f'A_LFsubsamp_exp{np.int(dur/60)}min'
    elif sampopt == 'B':
        sampopt_txt = 'B_LFfit'; N_hf = 2
        txtadd = f'B_LFfit_exp{np.int(dur/60)}min'
    elif sampopt == 'C':
        sampopt_txt = 'C_LFisHFfit'; N_hf = 2
        txtadd = f'C_LFisHFfit_exp{np.int(dur/60)}min'
    
    # Initialise
    plt.figure(10000+gg,figsize=(11,4))
    plt.figure(20000+gg,figsize=(5.5,4))
    
    # Loop over the different cases
    # for nn in np.array((0,1,2,3)):
    for nn in range(2):
        
        # Noise texts (figure titles and names)
        if nn == 0:
            indcase = 'GoodInd'
            indalias = f'GoodInd, {gprcase}'
        elif nn == 1:
            indcase = 'WorseInd'
            indalias = f'WorseInd, {gprcase}'
     
        # Load MPM and MPMU95 data
        if GPRvers == 'GPR':
            # header = ['NoIt','MPM_HFGPR','U95_MPM_HFGPR','SumVar_HFGPR']
            Dat_USMV = pd.read_csv(f'{in_files}/Conv_HF-GPR_{indcase}_Set-{HFkern}_{txtadd}_USMV_Nhf{N_hf}.csv', delimiter=',')
        elif GPRvers == 'MF-GPR':
            # header = ['NoIt','MPM_MFGPR','U95_MPM_MFGPR','SumVar_MFGPR']
            Dat_USMV = pd.read_csv(f'{in_files}/Conv_MF-GPR_{indcase}_Set-{MFkern1}-{MFkern2}_{txtadd}_USMV_Nhf{N_hf}.csv', delimiter=',')
    
        # Number of samples (not iterations), MPM and U95 of MPM    
        if GPRvers == 'GPR':
            nosamp_USMV = Dat_USMV['NoSamp'].to_numpy()
            MPM_USMV = Dat_USMV['MPM_HFGPR'].to_numpy() 
            MPMU95_USMV = Dat_USMV['U95_MPM_HFGPR'].to_numpy()
        else:
            nosamp_USMV = Dat_USMV['NoSamp'].to_numpy()
            MPM_USMV = Dat_USMV['MPM_MFGPR'].to_numpy() 
            MPMU95_USMV = Dat_USMV['U95_MPM_MFGPR'].to_numpy()
            
        # Max. number of available iterations
        itmaxall = count_USMV+1
        
        # Define the colormap Viridis and normalize it to the data range
        cmap = plt.get_cmap('jet') 
        norm = Normalize(vmin=0, vmax=itmaxall)
        
        # Create a ScalarMappable object to map data values to colors
        sm = cm.ScalarMappable(cmap=cmap, norm=norm)
        sm.set_array([]) 
        
        # Initialisation of assembled convergence criteria for different acquisition functions
        crit_rej = []
        crit_cov = []
        crit_cov_av = []
        crit_MAE = []
        crit_MAE_av = []
        crit_all = []
        
        # # Max. number of acquisition functions to analyse varies per noise case
        # # maxnumacq = 6 # not available anymore; IVRMV and IVRPD skipped
        # if GPRvers == 'GPR':
        #    acqcases = [0]
        # elif GPRvers == 'MF-GPR':
        #     acqcases = [0]
        
        # for mm in acqcases:
            
        # Initialise assembly of mean predictions
        ass_meanpred = []
        
        # Initialise rejection test for monotonicity
        RejectIt = np.ones(count_USMV+1)
             
        # initialise MAE 
        D_MAE = np.nan*np.zeros(count_USMV+1)
        D_MAE_av = np.nan*np.zeros(count_USMV+1)
        
        # Initialise mean, standard deviation and coefficient of variation of MPM
        stdconv_mpm = np.nan*np.zeros(count_USMV+1)
        meanconv_mpm = np.nan*np.zeros(count_USMV+1)
        cov_mpm = np.nan*np.zeros(count_USMV+1)
        cov_mpm_last = np.nan*np.zeros(count_USMV+1)
        
        # Initialise final criterion that combines 1,2,3 (starts at 0)
        critall_pa = np.ones(count_USMV+1)
        
        # Loop over number of iterations
        for ii in range(count_USMV+1):
    
            # Color range for increasing it. number
            colnow = [0,0,0.5+(ii/itmaxall)/2]
            
            # Load data - full predicted HF distributions
            if GPRvers == 'GPR':
                preddata = np.genfromtxt(f'{in_files}/HFGPR_HFpred_{indcase}_Set-{HFkern}-nf_{txtadd}_HFacq-USMV_Nhf{N_hf}_HFit{ii}.csv', skip_header=1, delimiter=',')
            elif GPRvers == 'MF-GPR':
                preddata = np.genfromtxt(f'{in_files}/MFGPR_HFpred_{indcase}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_MFacq-USMV_Nhf{N_hf}_MFit{ii}.csv', skip_header=1, delimiter=',')
            MPMs = MPM_USMV.copy()
            MPMU95s = MPMU95_USMV.copy() 
                           
            # Extract PoE range and mean predictions (HF version MF-GPR at same columns as GPR predictions)
            Prange_pred = preddata[:,0]
            mean_pred = preddata[:,1]
            
            # MPM and MPM U95 values up to the present iteration
            mpms_now = MPMs[0:ii]
            u95mpms_now = MPMU95s[0:ii]
            
            ##################  
            # Crit1: Reject non-monotonic results 
            ################## 
              
            # Check whether prediction in present iteration is monotonically decreasing, 
            # and that there is a difference between top and down of more than 1% of top value
            eps1 = 0.01*mean_pred[0]
            if all(mean_pred[i] > mean_pred[i+1] for i in range(len(mean_pred)-1)):
                # Yes, monotonically decreasing
                if mean_pred[0]-mean_pred[-1] > eps1:
                    # Mon. decreasing, but difference top-down big enough
                    RejectIt[ii] = 0
                else:
                    # Mon. decreasing, but difference top-down too small
                    RejectIt[ii] = 1
            else:
                # No, not monotonically decreasing
                RejectIt[ii] = 1
            
            ################## End crit1
    
            # Plot all distributions on top of each other
            if plotalldistroneplot == 1:
                plt.figure(100000*gg+2000+10*nn)
                if RejectIt[ii] == 1:
                    plt.plot(mean_pred,Prange_pred,color='grey')
                    plt.text(mean_pred[-1],Prange_pred[-1],f'{ii}',color='grey')                    
                else:
                    if np.mod(ii,20) == 0:
                        plt.plot(mean_pred,Prange_pred,color=sm.to_rgba(ii),label=f'Iteration {ii}')
                    else:
                        plt.plot(mean_pred,Prange_pred,color=sm.to_rgba(ii))
                    plt.text(mean_pred[-1],Prange_pred[-1],f'{ii}',color=sm.to_rgba(ii))
                        
            # Save these data to use in convergence assessment
            ass_meanpred.append(mean_pred)
            
            ##################  
            # Crit2 alt: Calculate maximum absolute error
            ##################     
                                
            # Loop starts from 2 available iterations.
            if ii > 0:
                
                # DNR present and previous iteration
                pred1 = ass_meanpred[ii-1]
                pred2 = ass_meanpred[ii]
                
                # max absolute difference
                # MAE = np.mean(np.abs(pred1-pred2))
                MAE = np.max(np.abs(pred1-pred2))
                # MAE = np.sqrt(np.mean((pred1-pred2)**2))
                
                # Assemble data
                D_MAE[ii] = MAE
                        
            # Calculate the MAE over the last N_mae iterations
            # Start at > 1 because D_MAE[:0] = [] and D_MAE[:1] = D_MAE[0] = nan
            if ii > 1:
                if ii <= N_mae-1:
                    # "Burn-in": discard portion of simulation results to ensure steady state (or treat them differently)
                    D_mae_t = D_MAE[:ii][np.isfinite(D_MAE[:ii])]
                elif ii > N_mae-1:
                    D_mae_t = D_MAE[ii-N_mae:ii][np.isfinite(D_MAE[ii-N_mae:ii])]
                D_MAE_av[ii] = sum(D_mae_t)/len(D_mae_t)
              
           
           ################## End crit2 alt
           
            ##################  
            # Crit3: Calculate standard deviation and COV of MPM over existing iterations
            ##################     
              
            # Sample standard deviation
            stdconv_mpm[ii] = np.std(mpms_now)
            
            # Sample mean
            meanconv_mpm[ii] = np.mean(mpms_now)  
            
            # Coefficient of variation up to present iteration
            cov_mpm[ii] = stdconv_mpm[ii]/meanconv_mpm[ii]
    
            # Very erratic; calculate the COV over the last N_cov iterations
            # Start at > 0 because mpms_now[:0] = []
            if ii > 1:
                # N_cov = 10
                if ii <= N_cov-1:
                    # "Burn-in": discard portion of simulation results to ensure steady state (or treat them differently)
                    mpms_t = mpms_now[:ii][np.isfinite(mpms_now[:ii])]
                elif ii > N_cov-1:
                    mpms_t = mpms_now[ii-N_cov:ii][np.isfinite(mpms_now[ii-N_cov:ii])]
                # cov_mpm_last[ii] = (np.std(mpms_t)/np.sqrt(len(mpms_t)))/np.mean(mpms_t)  
                cov_mpm_last[ii] = np.std(mpms_t)/np.mean(mpms_t)   
                # cov_mpm_last[ii] = np.std(mpms_t)
            
            ################## End crit3
                       
            
            
            ###########
            # One criterion that combines them all
            ###########
                
            # First check monotonicity criterion
            if RejectIt[ii] == 1:
                critall_pa[ii] = 0
           
            # Check the mean absolute error convergence criterion
            if D_MAE_av[ii] > lim_MAE_conv:
                critall_pa[ii] = 0
            elif np.isnan(D_MAE_av[ii]):
                critall_pa[ii] = 0  
                
            # Check the MPM / quantile COV convergence criterion
            if cov_mpm_last[ii] > lim_cov_conv:
                critall_pa[ii] = 0
            elif np.isnan(cov_mpm_last[ii]):
                critall_pa[ii] = 0
          
    
        # Finish figure of all distributions
        if plotalldistroneplot == 1:
            plt.figure(100000*gg+2000+10*nn)
            plt.yscale('log')
            plt.title(f'Indicator quality case: {indalias} \n {gprcase} (grey = rejected iteration)')
            plt.ylabel('P(c$\geq$C) [-]')
            plt.xlabel('C [m]')
            plt.ylim([4*10**(-4),2*10**(-1)]) # room for legend
            plt.xlim([7.5,12.5])
            plt.legend(ncol=2)
            plt.plot([0,15],[pexp,pexp],':',color='black')
            plt.text(12.1,pexp*1.2,'$P_{exp}$')
            plt.savefig(f'{out_figs}/AllDistrForIts_{indalias}_{gprcase}_VarAcq.pdf',bbox_inches='tight')
        
        # Save for all acquisition functions
        crit_rej.append(RejectIt)
        crit_cov.append(cov_mpm)
        crit_cov_av.append(cov_mpm_last)
        crit_MAE.append(D_MAE)
        crit_MAE_av.append(D_MAE_av)
        crit_all.append(critall_pa)
        
         
        ###########
        # First converged iteration number
        ###########
        if np.max(crit_all) == 0:
            ic_USMV = np.nan
        else:
            ic_USMV = np.argmax(np.squeeze(crit_all)>0)
        samp_conv = nosamp_USMV[ic_USMV]
     
        
        ###########
        # Some plot settings
        ###########
        
        # Location of text 'Criterion'
        x_loc_crit = 170
        x_ticks = np.arange(0,202,25)
        
        # Set the rejected iterations to NaN in the other criteria plots
        # nanmask_rej_USMV = np.where(crit_rej[1]==1,np.nan,1)
        nanmask_rej_USMV = np.where(np.squeeze(crit_rej)==1,np.nan,1)
        
        ###########
        # MPM plot for all available iterations
        ###########
        
       
        # Same but all in one plot
        plt.figure(10000+gg)
        plt.subplot(2,2,nn+1)
        plt.plot(nosamp_USMV,MPM_USMV*nanmask_rej_USMV,linestyle=linestyles[1],color=cols[1],label='USMV')
        plt.plot([0,seedmaxplot],[MPM_true_MCS,MPM_true_MCS],linestyle=':',color='grey',linewidth=0.8) # ,label='Truth (MCS)')
        plt.text(165,10.1,'MCS \'truth\'',color='grey')
        plt.xticks(ticks=x_ticks)
        plt.xlim([0,seedmaxplot])
        plt.xlabel('Number of samples $j$ [-]')
        plt.yticks(ticks=[9,9.5,10,10.5,11,11.5])
        plt.ylim([9.5,11])
        if nn == 1:
            plt.legend(ncol=4)
        if nn == 0:
            plt.ylabel('$\widehat{C\'\'}$ [m]')
        plt.title(indalias)
    
        # ###########
        # # Final plot including convergence criteria
        # ###########
     
        # Convergence criterion plot 
        plt.figure(20000+gg)
        
        # MPM    
        ax1 = plt.subplot(3,1,1)
        plt.ylabel('$\widehat{C\'\'}$ [m]')
        plt.plot(nosamp_USMV[:ic_USMV-1],MPM_USMV[:ic_USMV-1],linestyle=linestyles[nn],color=cols2[nn],label=indalias)  
        plt.plot(nosamp_USMV[ic_USMV:],MPM_USMV[ic_USMV:],linestyle=linestyles[nn],color=cols2[nn],alpha=0.4)  
        plt.plot(nosamp_USMV[ic_USMV],MPM_USMV[ic_USMV],linestyle='none',color=cols2[nn],marker = 'x') 
        if nn == 1:
            plt.plot([0,seedmaxplot],[MPM_true_MCS,MPM_true_MCS],linestyle=':',color='grey',linewidth=0.8,label='MCS \'truth\'')
            plt.yticks(ticks=[9,9.5,10,10.5,11,11.5])
            plt.ylim([9,11])
            plt.xlim([0,seedmaxplot])
            plt.xticks(ticks=x_ticks,labels=[])
            if GPRvers == 'GPR':
                plt.legend(ncol=3,loc='lower right')
        
        # MAE criterion 2alt
        ax2 = plt.subplot(3,1,2)
        plt.ylabel(r'$\overline{E}_{%i}$ [m]' % (N_mae))
        plt.plot(nosamp_USMV,np.squeeze(crit_MAE_av)*nanmask_rej_USMV,linestyle=linestyles[nn],color=cols2[nn],label=indalias)
        if nn == 1:
            plt.xlim([0,seedmaxplot])
            plt.plot([0,seedmaxplot],[lim_MAE_conv,lim_MAE_conv],linestyle='--',color='grey',linewidth=0.8)
            plt.text(x_loc_crit,lim_MAE_conv*1.5,'Criterion',color='grey')
            plt.yscale('log')
            plt.ylim([10**(-4),1])
            plt.xticks(ticks=x_ticks,labels=[])
               
        # MPM COV criterion 3 (plus averaged over a number of iterations)
        ax3 = plt.subplot(3,1,3)
        plt.ylabel(r'$C_{%i}$ [-]' % (N_cov))
        plt.plot(nosamp_USMV,np.squeeze(crit_cov_av)*nanmask_rej_USMV,linestyle=linestyles[nn],color=cols2[nn],label=indalias)
        if nn == 1:
            plt.xlim([0,seedmaxplot])
            plt.plot([0,seedmaxplot],[lim_cov_conv,lim_cov_conv],linestyle='--',color='grey',linewidth=0.8)
            plt.text(x_loc_crit,lim_cov_conv*1.5,'Criterion',color='grey')
            plt.yscale('log')
            plt.ylim([10**(-5),10**(-1)])
            plt.xticks(ticks=x_ticks)
            plt.xlabel('Number of samples $j$ [-]')
            
        # ###########
        # # Final plot including convergence criteria, GPR and MF-GPR in one plot
        # ###########           
        
        indleg = gg*2+nn
        
        
        # Convergence criterion plot 
        plt.figure(30000)
        
        # MPM    
        ax1 = plt.subplot(3,1,1)
        plt.ylabel('$\widehat{C\'\'}$ [m]')
        plt.plot(nosamp_USMV[:ic_USMV-1],MPM_USMV[:ic_USMV-1],linestyle=linestyles[indleg],color=cols2[indleg],label=indalias)  
        plt.plot(nosamp_USMV[ic_USMV:],MPM_USMV[ic_USMV:],linestyle=linestyles[indleg],color=cols2[indleg],alpha=0.4)  
        plt.plot(nosamp_USMV[ic_USMV],MPM_USMV[ic_USMV],linestyle='none',color=cols2[indleg],marker = 'x') 
        if indleg == 3:
            plt.plot([0,seedmaxplot],[MPM_true_MCS,MPM_true_MCS],linestyle=':',color='grey',linewidth=0.8,label='MCS \'truth\'')
            plt.yticks(ticks=[9,9.5,10,10.5,11,11.5])
            plt.ylim([9,11])
            plt.xlim([0,seedmaxplot])
            plt.xticks(ticks=x_ticks,labels=[])
            plt.legend(ncol=3,loc='lower right')
        
        # MAE criterion 2alt
        ax2 = plt.subplot(3,1,2)
        plt.ylabel(r'$\overline{E}_{%i}$ [m]' % (N_mae))
        plt.plot(nosamp_USMV,np.squeeze(crit_MAE_av)*nanmask_rej_USMV,linestyle=linestyles[indleg],color=cols2[indleg],label=indalias)
        if indleg == 3:
            plt.xlim([0,seedmaxplot])
            plt.plot([0,seedmaxplot],[lim_MAE_conv,lim_MAE_conv],linestyle='--',color='grey',linewidth=0.8)
            plt.text(x_loc_crit,lim_MAE_conv*1.5,'Criterion',color='grey')
            plt.yscale('log')
            plt.ylim([3*10**(-3),1])
            plt.xticks(ticks=x_ticks,labels=[])
               
        # MPM COV criterion 3 (plus averaged over a number of iterations)
        ax3 = plt.subplot(3,1,3)
        plt.ylabel(r'$C_{%i}$ [-]' % (N_cov))
        plt.plot(nosamp_USMV,np.squeeze(crit_cov_av)*nanmask_rej_USMV,linestyle=linestyles[indleg],color=cols2[indleg],label=indalias)
        if indleg == 3:
            plt.xlim([0,seedmaxplot])
            plt.plot([0,seedmaxplot],[lim_cov_conv,lim_cov_conv],linestyle='--',color='grey',linewidth=0.8)
            plt.text(x_loc_crit,lim_cov_conv*1.5,'Criterion',color='grey')
            plt.yscale('log')
            plt.ylim([10**(-5),10**(-1)])
            plt.xticks(ticks=x_ticks)
            plt.xlabel('Number of samples $j$ [-]')
        
        
        
        
        
        
            
        ###### Print results to screen
        print(f'FULL CONVERGENCE WITH PRESENT CRITERIA - {gprcase}:')
        print(f'Case {indalias}, USMV: converged at {ic_USMV} iterations - {ic_USMV+N_hf} samples (of max. {count_USMV})')
       
    # Finalise and save plots
    plt.figure(10000+gg)
    plt.savefig(f'{out_figs}/FigXX_Vary_acquisition_AllAcq_{gprcase}_Set-{HFkern}-nf_{sampopt_txt}_Nhf{N_hf}_MPM.pdf',bbox_inches='tight')
    
    plt.figure(20000+gg)
    plt.savefig(f'{out_figs}/FigXX_ConvergencePlots_{gprcase}_acq{acqsel}_Set-{HFkern}-nf_{sampopt_txt}_Nhf{N_hf}.pdf',bbox_inches='tight')
    
plt.figure(30000)
plt.savefig(f'{out_figs}/Fig13_ConvergencePlots_{gprcase}_acq{acqsel}_Set-{HFkern}-nf_{sampopt_txt}_Nhf{N_hf}_all.pdf',bbox_inches='tight')
    
       