# -*- coding: utf-8 -*-
"""
Created on Thu Aug 10 16:25:13 2023

@author: smvanessen

"""

# Clear all variables
from IPython import get_ipython
get_ipython().magic('reset -sf')

# Standard toolboxes
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import glob
import os
import re

# Add some other functions
from functions.subsample_routines import subsample_binned
from functions.distribution_scripts import MPM_from_DNR
from functions.theoretical_distributions import make_cdf_Forristall as Forr
from functions.distribution_scripts import calc_distribution

# Convergence criterion
from functions.convergenceAS import stopping_crit_AS

# Close all figures
plt.close('all')


#############
# START INPUT
#############

# Plotsize
fsize_h = 7
fsize_v = 4.5

# Number of HF samples to include in plot
sampmaxplot1 = 70.5 # Convergence
sampstep1 = 10
sampmaxplot2 = 60.5 # All considered samples
sampstep2 = 10

# Acquisition function type to evaluate.
acqnames = ['USMV']
# acqnames = ['MD']

# Select used initial sampling option. Note: you can only select options for 
# which you first generated data with the Adaptive Screening method. For a 
# description of options A, B and C, also see Adaptive Screening. 
# Options: 'A', 'B' or 'C'
sampopt = 'B'

# Folder where the input CSV files are (generated by the Adaptive Screening procedure) (in)
# Folder where we write the output figures (out)
in_files = 'out_files_waves'
out_figs = 'out_figs_paper'

# Corresponding line styles and colors in the plots
stylesgpr = ['-.'];     colsgpr = ['green']  # matching distribution plots
stylesmfgpr = ['--'];   colsmfgpr = ['blue'] # matching distribution plots

# Kernel name single-fidelity HF-GPR (preferred Mat32 for all)
HFkern = 'Mat32'
MFkern1 = 'Mat32' 
MFkern2 = 'Mat32'

# Spectral properties basis wave system (for ref. Forristall and Pexp)
dur = 3600 # target duration
Hs = 10 # sign wave height
Tp = 11 # peak wave period
wd = 30 # water depth
Vs = 0 # current speed or vessel speed
T1 = Tp/1.198 # estimated from relation for JONSWAP

# Number of seeds
Ns = 50

# Forristall
VrangeRF = np.linspace(0, 15, 101)[:, None]

# Annotation switch for the plot of the MPM as function of iterations until 
# convergence. Paper figure 12 includes annotation.
# 1 = with annotation, 0 = without annotation
annon = 1

# Switch for plotting of the intermediate results as in appendix
# 1 = plots of intermediate iterations, 0 = no plots of intermediate iterations
plotintermediateresults = 0

# SQI values for the different indicator quality cases (sqi), number of HF 
# samples for convergence with GPR (nev_gpr) and number of HF samples for 
# convergence with MF-GPR (nev_mfgpr). 
# Cases: pi = PerfInd, gi = GoodInd, mi = MediumInd and bi = BadInd
sqi_gi = 1.64
sqi_bi = 3.07

# Lower and upper limit for event duration CFD [s]
tlow = 10
thigh = 50

# Convergence limits stopping criterion
limMAE = 0.05 # [m]
limCOV = 0.003



#############
# END INPUT
#############

# Reference conventional result case (read from file)
convset = 'fitted'
filenameCONV = f'inputfiles_waves/Conventional_MPM_U95ofMPM_MPMfit_U95ofMPMfit.txt'
fconv = open(filenameCONV,'r')
fdata = fconv.read()
convresults = list(fdata.split(","))

if convset == 'fitted':
    MPM_conv = float(convresults[2])
    MPMU95_conv = float(convresults[3])
elif convset == 'unfitted':
    MPM_conv = float(convresults[0])
    MPMU95_conv = float(convresults[1])
HFtime_conv = 5


# Check wether output directory exist, if not make it
isExist1 = os.path.exists(out_figs)
if not isExist1:
   os.makedirs(out_figs)

# Initialise
plt.figure(10000,figsize=(12.5,3))
plt.figure(20000,figsize=(12.5,3))
x_ticks = np.arange(0,202,25)

# Initial sample options (see Adaptive Screening)
if sampopt == 'A':
    sampopt_txt = 'A_LFsubsamp'; N_hf = 2
    txtadd = f'A_LFsubsamp_exp{np.int(dur/60)}min'
elif sampopt == 'B':
    sampopt_txt = 'B_LFfit'; N_hf = 2
    txtadd = f'B_LFfit_exp{np.int(dur/60)}min'
elif sampopt == 'C':
    sampopt_txt = 'C_LFisHFfit'; N_hf = 2
    txtadd = f'C_LFisHFfit_exp{np.int(dur/60)}min'

# Derive target probability of exceedance, using the average number of wave encounters in target duration
ne_exp = dur/Tp 
pexp = 1/ne_exp

# Calc Forristall ref distribution
p_For,dum2 = Forr(VrangeRF,Hs,T1,wd,Vs)

# Initialise
conv_mpm_gpr = np.nan*np.zeros((2,len(acqnames)))
conv_mpmu95_gpr = np.nan*np.zeros((2,len(acqnames)))
conv_mpm_mfgpr = np.nan*np.zeros((2,len(acqnames)))
conv_mpmu95_mfgpr = np.nan*np.zeros((2,len(acqnames)))
NumSampConv_gpr = np.nan*np.zeros((2,len(acqnames)))
NumSampConv_mfgpr = np.nan*np.zeros((2,len(acqnames)))

# Initialise plots as function of SQI (for selected acquisition function)
sqi_all = []
mpm_gpr_all = [];        mpm_mfgpr_all = []
mpmu95_gpr_all = [];     mpmu95_mfgpr_all = []
time_gpr_low = [];       time_mfgpr_low =[]
time_gpr_hig = [];       time_mfgpr_hig = []

# Loop over conditions
for nn in range(2):
    
    # Noise texts (figure titles and names)
    if nn == 0:
        indcase = 'GoodInd'
        indalias = 'C_{good}\prime'
        sqi = sqi_gi
        # nev_gpr = nev_gpr_gi
        # nev_mfgpr = nev_mfgpr_gi
    elif nn == 1:
        indcase = 'WorseInd'
        indalias = 'C_{worse}\prime'
        sqi = sqi_bi
        # nev_gpr = nev_gpr_bi
        # nev_mfgpr = nev_mfgpr_bi

    # Append SQI values and number of HF events
    sqi_all = np.append(sqi_all,sqi)    
    
    # Read MCS LF and HF true and training time traces from CSV file   
    # ALREADY CORRECTED FOR BASIS NUMBER OF WAVES, INSTEAD OF NUMBER OF LF OR HF PEAKS, WITHIN ADAPTIVE SCREENING SCRIPT
    DatLFtrue = np.loadtxt(f'{in_files}/TrueLFdata_{indcase}_{txtadd}.csv', delimiter=',')
    v_true_l = DatLFtrue[:,0]
    p_true_l = DatLFtrue[:,1]
    DatHFtrue = np.loadtxt(f'{in_files}/TrueHFdata_{indcase}_{txtadd}.csv', delimiter=',') # 50h MCS realisation used in Adaptive Screening
    v_true_h = DatHFtrue[:,0]
    p_true_h = DatHFtrue[:,1]
    DatLFtrain = np.loadtxt(f'{in_files}/TrainLFdata_{indcase}_{txtadd}.csv', delimiter=',')
    v_train_l = DatLFtrain[:,0]
    p_train_l = DatLFtrain[:,1]

    # Subsample the MCS true distribution data for plotting later
    # binsize = 0.05
    binsize = 0.02
    bins = np.arange(0,15,binsize)
    v_true_l_sampled,p_true_l_sampled = subsample_binned(v_true_l,p_true_l,bins)
    v_true_h_sampled,p_true_h_sampled = subsample_binned(v_true_h,p_true_h,bins)
    
    # Max values (for plotting) 
    maxVal = np.ceil(1.1*np.nanmax(v_true_h)) # HF
    maxVal_lf = np.ceil(1.1*np.nanmax(v_true_l)) # LF
    
    # Min PoE value (for plotting)
    minP = 0.9*np.nanmin(p_true_h) # HF
    
    # Calculate / derive true MPM for 50h MCS realisation used in Adaptive Screening
    MPM_true_MCS = MPM_from_DNR(v_true_h,p_true_h,ne_exp) 
    
    # Check of all available 50h realisations
    MPM_true_MCS_50hseeds = np.zeros(10)
    for bb in range(10): 
        
        # Read data other seeds
        DatHFtrue_50hseeds = np.loadtxt(f'inputfiles_waves/Lin2nd_trueHFdata_real{bb}_50seeds.csv', delimiter=',')
        # tpeaks_true_h_50hseeds = DatHFtrue_50hseeds[:,0]
        peaks_true_h_50hseeds = DatHFtrue_50hseeds[:,1]
        
        # Calculate distributions
        v_true_h_50hseeds,p_true_h_50hseeds_temp,_ = calc_distribution(peaks_true_h_50hseeds)
        
        # Correct for using one estimated number of waves as basis instead of the number 
        # of data points. 
        Tdata = dur*Ns # Total duration [s]
        Nwaves = Tdata/dur*ne_exp # Estimated number of wave encounters in full experimental duration

        # Correct for basis Nwaves instead of the number of LF or HF peaks
        N_true_h_50hseeds = len(v_true_h_50hseeds)
        p_true_h_50hseeds = p_true_h_50hseeds_temp*N_true_h_50hseeds/Nwaves
       
        # Calculate MPM        
        MPM_true_MCS_50hseeds[bb] = MPM_from_DNR(v_true_h_50hseeds,p_true_h_50hseeds,ne_exp) 
        
        # Delete data of extra seeds
        del peaks_true_h_50hseeds,v_true_h_50hseeds,p_true_h_50hseeds
    
    # U95 over all 10 times a 50h realisation
    MPM_true_MCS_50hseeds_U95 = 1.96*np.std(MPM_true_MCS_50hseeds)    
    
    
    print('Used true HF 50h realisation:')
    print(f'{MPM_true_MCS}m')
    print('True HF mean and U95 over ten 50h realisations:')
    print(f'{np.mean(MPM_true_MCS_50hseeds)}m and U95 {MPM_true_MCS_50hseeds_U95}m')
    
    # Calculate / derive Forristall reference MPM
    if indcase == 'GoodInd':
        MPM_true_For = MPM_from_DNR(VrangeRF,p_For,ne_exp)
    else:
        MPM_true_For = np.nan

        
    # Loop over selected acquisition functions (only functions that are available for both GPR and MF-GPR)
    for mm in range(len(acqnames)):
        acqname = acqnames[mm]
        print(f'Case {indcase} ({nn}/2), acq. function {acqname} ({mm}/{len(acqnames)-1})')
        
        # Load data from HF-GPR and MF-GPR with all iteration results (MPM etc.)
        file_conv_hf = f'{in_files}/Conv_HF-GPR_{indcase}_Set-{HFkern}_{txtadd}_{acqname}_Nhf{N_hf}.csv'
        file_conv_mf = f'{in_files}/Conv_MF-GPR_{indcase}_Set-{MFkern1}-{MFkern2}_{txtadd}_{acqname}_Nhf{N_hf}.csv'
        
        dat_conv_hf_all = pd.read_csv(file_conv_hf, delimiter=',')
        dat_conv_mf_all = pd.read_csv(file_conv_mf, delimiter=',')
        
        # Extract last iteration number and max. number of samples
        nosamp_hf = dat_conv_hf_all['NoSamp'].to_numpy()
        nosamp_mf = dat_conv_mf_all['NoSamp'].to_numpy()
        
        # Max. number of samples, and max. number of iterations
        maxsamp_hf = int(np.nanmax(nosamp_hf))
        maxsamp_mf = int(np.nanmax(nosamp_mf))
        maxit_hf = int(maxsamp_hf-N_hf)
        maxit_mf = int(maxsamp_mf-N_hf)
        
        print(f'{indcase} with HF-GPR and {acqname}: available are {maxit_hf} iterations ({maxsamp_hf} samples)')
        print(f'{indcase} with MF-GPR and {acqname}: available are {maxit_mf} iterations ({maxsamp_mf} samples)')
        
        
        #### Load data for this last iteration and plot convergence
      
        # Actual MPM values
        mpms_GPR = dat_conv_hf_all['MPM_HFGPR'].to_numpy()
        mpmsU95_GPR = dat_conv_hf_all['U95_MPM_HFGPR'].to_numpy()
        
        try:        
            mpms_MFGPR = dat_conv_mf_all['MPM_MFGPR'].to_numpy() # Use when AS is run again after fixing typo
            mpmsU95_MFGPR = dat_conv_mf_all['U95_MPM_MFGPR'].to_numpy() # Use when AS is run again after fixing typo
        except:
            mpms_MFGPR = dat_conv_mf_all['MPM_HFGPR'].to_numpy() # Typo in earlier version Adaptive Screening; fixed, but present files still use this
            mpmsU95_MFGPR = dat_conv_mf_all['U95_MPM_HFGPR'].to_numpy() # Typo in earlier version Adaptive Screening; fixed, but present files still use this
        
        # Plot all iterations
        plt.figure(20000)
        plt.subplot(1,2,nn+1) 
        if nn == 0:
            plt.ylabel('$\widehat{C\'\'}$ [m]')
        plt.xlabel('Number of samples $j$ [-]')
        plt.xticks(ticks=x_ticks)
        plt.title(r'Indicator: $%s$' % indalias)
        plt.plot(nosamp_hf,mpms_GPR,linestyle=stylesgpr[mm],color=colsgpr[mm],label='GPR')  
        plt.plot(nosamp_mf,mpms_MFGPR,linestyle=stylesmfgpr[mm],color=colsmfgpr[mm],label='MF-GPR')    
        plt.fill_between(nosamp_hf.flatten(), (mpms_GPR - 0.5*mpmsU95_GPR).flatten(), 
                          (mpms_GPR + 0.5*mpmsU95_GPR).flatten(), facecolor=colsgpr[0], alpha=0.3,label='GPR U95%')
        plt.fill_between(nosamp_mf.flatten(), (mpms_MFGPR - 0.5*mpmsU95_MFGPR).flatten(), 
                          (mpms_MFGPR + 0.5*mpmsU95_MFGPR).flatten(), facecolor=colsmfgpr[0], alpha=0.3,label='MF-GPR U95%')
        plt.plot([0,sampmaxplot2],[MPM_true_MCS,MPM_true_MCS],linestyle=':',color='black'); plt.text(sampmaxplot2-30,MPM_true_MCS+0.03,'True MCS',color='black')
        plt.ylim([9,12])
        plt.xlim([N_hf-1,sampmaxplot2])
        plt.xticks(ticks=np.arange(0,sampmaxplot2,sampstep2,dtype=int),label=[])
        if nn == 0:
            plt.legend(ncol=2)
        
        
        #### Apply stopping criterion to determine at which iteration we are converged
        
        # File name start of CSV with predicted data per iteration (exclude "itXX.csv", will be added later)
        fn_preds_GPR = f'{in_files}/HFGPR_HFpred_{indcase}_Set-{HFkern}-nf_{txtadd}_HFacq-{acqname}_Nhf{N_hf}_HF'
        casename_GPR = f'{indcase} - GPR - {acqname}'
        min_it1 = 5
        ploton1 = False
        convit1,convsamp1 = stopping_crit_AS(nosamp_hf,mpms_GPR,mpmsU95_GPR,fn_preds_GPR,casename_GPR,pexp,min_it1,limMAE,limCOV,ploton1) # final
        
        fn_preds_MFGPR = f'{in_files}/MFGPR_HFpred_{indcase}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_MFacq-{acqname}_Nhf{N_hf}_MF'
        casename_MFGPR = f'{indcase} - MF-GPR - {acqname}'
        min_it2 = 5
        ploton2 = False
        convit2,convsamp2 = stopping_crit_AS(nosamp_mf,mpms_MFGPR,mpmsU95_MFGPR,fn_preds_MFGPR,casename_MFGPR,pexp,min_it2,limMAE,limCOV,ploton2) # final
        
        # Derive range in required duration [h]
        HFtime_gpr = [convsamp1*tlow/3600,convsamp1*thigh/3600]
        HFtime_mfgpr = [convsamp2*tlow/3600,convsamp2*thigh/3600]    
        
        time_gpr_low = np.append(time_gpr_low,HFtime_gpr[0])
        time_mfgpr_low = np.append(time_mfgpr_low,HFtime_mfgpr[0])
        time_gpr_hig = np.append(time_gpr_hig,HFtime_gpr[1])
        time_mfgpr_hig = np.append(time_mfgpr_hig,HFtime_mfgpr[1])

        # Prepare for annotation
        if annon == 1:
            i_samp10_hf = 10-N_hf
            gpr10 = np.round((mpms_GPR[i_samp10_hf]/MPM_true_MCS-1)*100,1);  signgpr10 = ''
            if gpr10>0: 
                signgpr10 = '+' 
            mfgpr10 = np.round((mpms_MFGPR[i_samp10_hf]/MPM_true_MCS-1)*100,1); signmfgpr10 = ''
            if mfgpr10>0: 
                signmfgpr10 = '+' 
            gprconv = np.round((mpms_GPR[convsamp1]/MPM_true_MCS-1)*100,1); signgprconv = ''
            if gprconv>0: 
                signgprconv = '+' 
            mfgprconv = np.round((mpms_MFGPR[convsamp2]/MPM_true_MCS-1)*100,1); signmfgprconv = ''
            if mfgprconv>0: 
                signmfgprconv = '+'  
            
        # Same but all in one plot
        plt.figure(10000)
        plt.subplot(1,2,nn+1) 
        plt.plot([0,sampmaxplot1],0.98*np.array([MPM_true_MCS,MPM_true_MCS]),linestyle=':',color='lightgrey',linewidth=0.3)
        plt.plot([0,sampmaxplot1],1.02*np.array([MPM_true_MCS,MPM_true_MCS]),linestyle=':',color='lightgrey',linewidth=0.3)
        plt.plot([0,sampmaxplot1],0.96*np.array([MPM_true_MCS,MPM_true_MCS]),linestyle=':',color='lightgrey',linewidth=0.3)
        plt.plot([0,sampmaxplot1],1.04*np.array([MPM_true_MCS,MPM_true_MCS]),linestyle=':',color='lightgrey',linewidth=0.3)
        if nn == 0:
            plt.ylabel('$\widehat{C\'\'}$ [m]')
            plt.text(0.93*sampmaxplot1,0.98*MPM_true_MCS+0.03,'-2%',color='lightgrey',fontsize=8)
            plt.text(0.93*sampmaxplot1,1.02*MPM_true_MCS+0.03,'+2%',color='lightgrey',fontsize=8)
            plt.text(0.93*sampmaxplot1,0.96*MPM_true_MCS+0.03,'-4%',color='lightgrey',fontsize=8)
            plt.text(0.93*sampmaxplot1,1.04*MPM_true_MCS+0.03,'+4%',color='lightgrey',fontsize=8)
        plt.xlabel('Number of samples $j$ [-]')
        plt.xticks(ticks=x_ticks)
        plt.title(r'Indicator: $%s$' % indalias)
        plt.plot(nosamp_hf[:convsamp1-1],mpms_GPR[:convsamp1-1],linestyle=stylesgpr[mm],color=colsgpr[mm],label='GPR')  
        plt.plot(nosamp_mf[:convsamp2-1],mpms_MFGPR[:convsamp2-1],linestyle=stylesmfgpr[mm],color=colsmfgpr[mm],label='MF-GPR')    
        plt.fill_between(nosamp_hf[:convsamp1-1].flatten(), (mpms_GPR[:convsamp1-1] - 0.5*mpmsU95_GPR[:convsamp1-1]).flatten(), 
                          (mpms_GPR[:convsamp1-1] + 0.5*mpmsU95_GPR[:convsamp1-1]).flatten(), facecolor=colsgpr[0], alpha=0.3,label='GPR U95%')
        plt.fill_between(nosamp_mf[:convsamp2-1].flatten(), (mpms_MFGPR[:convsamp2-1] - 0.5*mpmsU95_MFGPR[:convsamp2-1]).flatten(), 
                          (mpms_MFGPR[:convsamp2-1] + 0.5*mpmsU95_MFGPR[:convsamp2-1]).flatten(), facecolor=colsmfgpr[0], alpha=0.3,label='MF-GPR U95%')
        # plt.xticks(np.arange(0,sampmaxplot,sampstep,dtype=int))
        plt.plot([0,sampmaxplot1],[MPM_true_MCS,MPM_true_MCS],linestyle=':',color='black'); plt.text(0.85*sampmaxplot1,MPM_true_MCS+0.03,'True MCS',color='black')
        plt.ylim([9,12])
        plt.xlim([N_hf-1,sampmaxplot1])
        plt.xticks(ticks=np.arange(0,sampmaxplot1,sampstep1,dtype=int),label=[])
        if nn == 0:
            plt.legend(ncol=2)
        if annon == 1:
            plt.plot([nosamp_hf[i_samp10_hf],nosamp_hf[i_samp10_hf]],[MPM_true_MCS,11.8],color='grey',linewidth=0.2)
            plt.text(nosamp_hf[i_samp10_hf]+1,11.5,f'10 samples\nGPR: {signgpr10}{gpr10}%\nMF-GPR: {signmfgpr10}{mfgpr10}%',color='grey',fontsize=8)
            plt.plot([nosamp_hf[convsamp1-1]-1,nosamp_hf[convsamp1-1]-1],[mpms_GPR[convsamp1-1],11.05],color='green',linewidth=0.2)
            plt.text(nosamp_hf[convsamp1-1],10.9,f'converged\nGPR: {signgprconv}{gprconv}%',color='green',fontsize=8)
            plt.plot([nosamp_mf[convsamp2-1]-1,nosamp_mf[convsamp2-1]-1],[mpms_MFGPR[convsamp2-1],9.25],color='blue',linewidth=0.2)
            plt.text(nosamp_mf[convsamp2-1],9.2,f'converged\nMF-GPR: {signmfgprconv}{mfgprconv}%',color='blue',fontsize=8)
            
        #### CALCULATE AND PRINT MPM VALUES AND U95% FOR CONVERGED RESULTS
        conv_mpm_gpr[nn,mm] = mpms_GPR[convsamp1]
        conv_mpmu95_gpr[nn,mm] = mpmsU95_GPR[convsamp1]
        conv_mpm_mfgpr[nn,mm] = mpms_MFGPR[convsamp2]
        conv_mpmu95_mfgpr[nn,mm] = mpmsU95_MFGPR[convsamp2]
        
        # Assemble number of samples for convergence
        NumSampConv_gpr[nn,mm] = convsamp1
        NumSampConv_mfgpr[nn,mm] = convsamp2
        
        print(f'Case: {indcase}, {dur/3600}h:')
        print(f'MPM True MCS: {MPM_true_MCS} m')
        print(f'MPM True Forristall: {MPM_true_For} m')
        print(f'MPM GPR {acqname}: {conv_mpm_gpr[nn,mm]} m (w.r.t. MCS {(conv_mpm_gpr[nn,mm]/MPM_true_MCS-1)*100}%), U95 {conv_mpmu95_gpr[nn,mm]} m')
        print(f'MPM MF-GPR {acqname}: {conv_mpm_mfgpr[nn,mm]} m (w.r.t. MCS {(conv_mpm_mfgpr[nn,mm]/MPM_true_MCS-1)*100}%), U95 {conv_mpmu95_mfgpr[nn,mm]} m')
        print(f'MPM reference conventional: {MPM_conv} m (w.r.t. MCS {(MPM_conv/MPM_true_MCS-1)*100}%), U95 {MPMU95_conv} m')
        
        ############### READ AND PLOT CONVERGED GPR RESULTS FOR PRESENT CASE & ACQ FUNCT
        
        # GPR (2x)
        DatHFtrainIt1 = np.genfromtxt(f'{in_files}/HFGPR_TrainData_{indcase}_Set-{HFkern}-nf_{txtadd}_HFacq-{acqname}_Nhf{N_hf}_HFit{convit1}.csv', skip_header=1, delimiter=',')
        v_train_h1 = DatHFtrainIt1[:,0]
        p_train_h1 = DatHFtrainIt1[:,1]
        
        DatHFpredIt1 = np.genfromtxt(f'{in_files}/HFGPR_HFpred_{indcase}_Set-{HFkern}-nf_{txtadd}_HFacq-{acqname}_Nhf{N_hf}_HFit{convit1}.csv', skip_header=1, delimiter=',')
        Prange_predHF = DatHFpredIt1[:,0]
        mean_high_gp_model = DatHFpredIt1[:,1]
        var_high_gp_model = DatHFpredIt1[:,2]
        U95_high_gp_model = 1.96*np.sqrt(var_high_gp_model)
        
        ###### Plot HF
        # Plot the posterior mean and variance (U95) from the reference HF gaussian process
        plt.figure(1000+nn,figsize=(fsize_h,fsize_v))
        plt.yscale('log')
        plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" MCS')
        plt.plot(v_train_h1,p_train_h1, color='r', marker='d',linestyle='None', label=f'{convsamp1} HF samples')
        plt.plot(mean_high_gp_model,Prange_predHF,'-.',color='g',label='GPR prediction')
        plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_gp_model - 0.5*U95_high_gp_model).flatten(), 
                          (mean_high_gp_model + 0.5*U95_high_gp_model).flatten(), facecolor='g', alpha=0.1,label='GPR prediction U95%')
        plt.plot([0,maxVal],[pexp,pexp],':',color='black')
        plt.text(0.9*maxVal,pexp*1.4,'$P_{exp}$')
        plt.ylabel('P(C\'\'$\geq$c) [-]')
        plt.xlabel('c [m]')
        plt.ylim([minP,1])
        plt.xlim([0,maxVal])
        plt.legend()
        plt.savefig(f'{out_figs}/Fig6_Prediction_HFGPR_{indcase}_Set-{HFkern}_{sampopt_txt}_{acqname}_{convsamp1}HFsamps.pdf',bbox_inches='tight')
       
        
        
        ############### READ AND PLOT CONVERGED MF-GPR RESULTS FOR PRESENT CASE & ACQ FUNCT
        
        # MF-GPR (4x)
        DatLFtrainIt2 = np.genfromtxt(f'{in_files}/MFGPR_LFTrainData_{indcase}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_MFacq-{acqname}_Nhf{N_hf}_MFit{convit2}.csv', skip_header=1, delimiter=',')
        v_train_l2 = DatLFtrainIt2[:,0]
        p_train_l2 = DatLFtrainIt2[:,1]
        
        DatHFtrainIt2 = np.genfromtxt(f'{in_files}/MFGPR_HFTrainData_{indcase}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_MFacq-{acqname}_Nhf{N_hf}_MFit{convit2}.csv', skip_header=1, delimiter=',')
        v_train_h2 = DatHFtrainIt2[:,0]
        p_train_h2 = DatHFtrainIt2[:,1]
                
        DatLFpredIt2 = np.genfromtxt(f'{in_files}/MFGPR_LFpred_{indcase}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_MFacq-{acqname}_Nhf{N_hf}_MFit{convit2}.csv', skip_header=1, delimiter=',')
        Prange_predLF = DatLFpredIt2[:,0]
        mean_low_mfgp_model = DatLFpredIt2[:,1]
        var_low_mfgp_model = DatLFpredIt2[:,2]
        U95_low_mfgp_model = 1.96*np.sqrt(var_low_mfgp_model)
        
        DatHFpredIt2 = np.genfromtxt(f'{in_files}/MFGPR_HFpred_{indcase}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_MFacq-{acqname}_Nhf{N_hf}_MFit{convit2}.csv', skip_header=1, delimiter=',')
        Prange_predHF = DatHFpredIt2[:,0]
        mean_high_mfgp_model = DatHFpredIt2[:,1]
        var_high_mfgp_model = DatHFpredIt2[:,2]
        U95_high_mfgp_model = 1.96*np.sqrt(var_high_mfgp_model)
        
        # Debugging
        plt.figure()
        plt.yscale('log')
        plt.plot(v_true_l_sampled,p_true_l_sampled,marker='.',color=(0.7,0.7,0.7),linestyle='None',label='LF "true" MCS')
        plt.plot(v_train_l2,p_train_l2, color='grey', marker='o',linestyle='None', label='LF samples')
        
        ###### Plot MF
        # Plot the posterior mean and variance (U95) from the MF gaussian process
        plt.figure(2000+nn,figsize=(fsize_h,fsize_v))
        plt.yscale('log')
        plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" MCS (validation)')
        plt.plot(v_train_l2,p_train_l2, color='grey', marker='o',linestyle='None', label='LF samples')
        plt.plot(v_train_h2,p_train_h2, color='r', marker='d',linestyle='None', label=f'{convsamp2} HF samples')
        plt.plot(mean_low_mfgp_model,Prange_predLF,'--', color='b',label='MF-GPR predictions')
        plt.fill_betweenx(Prange_predLF.flatten(), (mean_low_mfgp_model - 0.5*U95_low_mfgp_model).flatten(), 
                          (mean_low_mfgp_model + 0.5*U95_low_mfgp_model).flatten(), facecolor='b', alpha=0.3,label='MF-GPR predictions U95%')
        plt.plot(mean_high_mfgp_model,Prange_predHF, '--', color='b',label='MF-GPR prediction HF (label not used)')
        plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_mfgp_model - 0.5*U95_high_mfgp_model).flatten(), 
                          (mean_high_mfgp_model + 0.5*U95_high_mfgp_model).flatten(), facecolor='b', alpha=0.3,label='MF-GPR prediction HF U95% (label not printed)')
        plt.plot([0,maxVal],[pexp,pexp],':',color='black')
        plt.text(0.9*maxVal,pexp*1.4,'$P_{exp}$')
        plt.ylabel('P(C\'\'$\geq$c) [-]')
        plt.xlabel('c [m]')
        plt.ylim([minP,1])
        plt.xlim([0,maxVal])
        handles, labels = plt.gca().get_legend_handles_labels() # get handles and labels (in order to change legend order)
        order = [0,1,2,3,4] #specify order of items in legend
        plt.legend([handles[idx] for idx in order],[labels[idx] for idx in order]) #add legend to plot
        plt.savefig(f'{out_figs}/Fig6_Prediction_MFGPR_{indcase}_Set-{MFkern1}-{MFkern2}_{sampopt_txt}_{acqname}_{convsamp2}HFsamps1.pdf',bbox_inches='tight')
              
        # Alternative side-by-side plot
        plt.figure(2010+nn,figsize=(fsize_h*1.7,fsize_v*0.8))
        # -- subplot 2 (HF data)
        plt.subplot(1,2,2)
        plt.yscale('log') 
        plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" MCS (validation)')
        plt.plot(v_train_h2,p_train_h2, color='r', marker='d',linestyle='None',markersize=4, label=f'{convsamp2} HF samples')
        plt.plot(mean_high_mfgp_model,Prange_predHF, '--', color='b',label='MF-GPR pred. HF')
        plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_mfgp_model - 0.5*U95_high_mfgp_model).flatten(), 
                         (mean_high_mfgp_model + 0.5*U95_high_mfgp_model).flatten(), facecolor='b', alpha=0.1,label='MF-GPR pred. HF U95%')
        plt.plot([0,maxVal],[pexp,pexp],':',color='black') #,label='$P_{exp}$')
        plt.text(0.9*maxVal,pexp*1.4,'$P_{exp}$')
        plt.ylabel('P(C\'\'$\geq$c) [-]')
        plt.xlabel('c [m]')
        plt.xlim(left=0,right=maxVal)
        plt.ylim([minP,1])
        plt.legend(loc='best')
        # -- subplot1 (LF data)
        plt.subplot(1,2,1)
        plt.yscale('log') 
        plt.plot(v_train_l2,p_train_l2, color='grey', marker='o',linestyle='None',markersize=4, label='LF samples')
        plt.plot(mean_low_mfgp_model,Prange_predLF, '--', color='b',alpha=0.4,label='MF-GPR pred. LF')
        plt.fill_betweenx(Prange_predLF.flatten(), (mean_low_mfgp_model - 0.5*U95_low_mfgp_model).flatten(), 
                          (mean_low_mfgp_model + 0.5*U95_low_mfgp_model).flatten(), facecolor='b', alpha=0.1,label='MF-GPR pred. LF U95%')
        plt.ylabel('P(C\'$\geq$c) [-]')
        plt.xlabel('c [m]')
        plt.xlim(left=0,right=maxVal_lf)
        # -- y-axis
        plt.ylim([minP,1])
        plt.legend(loc='best')
        plt.savefig(f'{out_figs}/FigXX_Prediction_MFGPR_{indcase}_Set-{MFkern1}-{MFkern2}_{sampopt_txt}_{acqname}_{convsamp2}HFsamps2.pdf',bbox_inches='tight')
                
          
        ############### PLOT GPR AND MF-GPR IN SAME PLOT
        
        # ALL IN ONE Plot the posterior mean and variance (U95) from the MF gaussian process (without training data, as these are different for HF-GPR and MF-GPR)
        plt.figure(3000+nn,figsize=(fsize_h,fsize_v))
        plt.yscale('log')
        plt.plot(v_true_l_sampled,p_true_l_sampled,marker='.',color=(0.7,0.7,0.7),linestyle='None',label='LF "true" MCS')
        plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" MCS') 
        plt.plot(mean_low_mfgp_model,Prange_predLF, '--', color='b',label='MF-GPR predictions')
        plt.fill_betweenx(Prange_predLF.flatten(), (mean_low_mfgp_model - 0.5*U95_low_mfgp_model).flatten(), 
                          (mean_low_mfgp_model + 0.5*U95_low_mfgp_model).flatten(), facecolor='b', alpha=0.3,label='MF-GPR predictions U95%')
        plt.plot(mean_high_mfgp_model,Prange_predHF, '--', color='b',label='MF-GPR prediction HF  (label not printed)')
        plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_mfgp_model - 0.5*U95_high_mfgp_model).flatten(), 
                          (mean_high_mfgp_model + 0.5*U95_high_mfgp_model).flatten(), facecolor='b', alpha=0.3,label='MF-GPR prediction HF U95% (label not printed)')
        plt.plot(mean_high_gp_model,Prange_predHF,'-.',color='g',label='GPR prediction HF')
        plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_gp_model - 0.5*U95_high_gp_model).flatten(), 
                          (mean_high_gp_model + 0.5*U95_high_gp_model).flatten(), facecolor='g', alpha=0.1,label='GPR prediction HF U95%')
        plt.plot([0,maxVal],[pexp,pexp],':',color='black')
        plt.text(0.9*maxVal,pexp*1.4,'$P_{exp}$')
        plt.ylabel('P(C\'\'$\geq$c) [-]')
        plt.xlabel('c [m]')
        plt.ylim([minP,1])
        plt.xlim([0,maxVal])
        handles, labels = plt.gca().get_legend_handles_labels() # get handles and labels (in order to change legend order)
        order = [0,1,2,3,6,7] #specify order of items in legend
        plt.legend([handles[idx] for idx in order],[labels[idx] for idx in order]) #add legend to plot
        plt.savefig(f'{out_figs}/FigXX_AllPred_{indcase}_HFSet-{HFkern}_MFSet-{MFkern1}-{MFkern2}_HF-{sampopt_txt}_MF-{sampopt_txt}_HFacq-{acqname}_MFacq-{acqname}_HFit{convit1}_MFit{convit2}.pdf',bbox_inches='tight')
        
        
        
    ############### PLOT DATA AS FUNCTION OF SQI, FOR SELECTED ACQUISITION FUNCTION
    
    # Save data as a function of SQI        
    acqsel = 0
    mpm_gpr_all = np.append(mpm_gpr_all,conv_mpm_gpr[nn,acqsel])
    mpm_mfgpr_all = np.append(mpm_mfgpr_all,conv_mpm_mfgpr[nn,acqsel])
    mpmu95_gpr_all = np.append(mpmu95_gpr_all,conv_mpmu95_gpr[nn,acqsel])
    mpmu95_mfgpr_all = np.append(mpmu95_mfgpr_all,conv_mpmu95_mfgpr[nn,acqsel])

    


############### Finalise and save convergence plot
plt.figure(10000)
if annon == 1:
    plt.savefig(f'{out_figs}/Fig5_Convergence_AllCases_GPR-{sampopt_txt}_MFGPR-{sampopt_txt}_{acqname}_converged_ann.pdf',bbox_inches='tight')
else:
    plt.savefig(f'{out_figs}/FigXX_Convergence_AllCases_GPR-{sampopt_txt}_MFGPR-{sampopt_txt}_{acqname}_converged.pdf',bbox_inches='tight')

            

# ############### Plot results as a function of SQI

# nev_gpr_all = np.squeeze(NumSampConv_gpr[:,acqsel])
# nev_mfgpr_all = np.squeeze(NumSampConv_mfgpr[:,acqsel])

# y_ticks1 = np.arange(8,15,1)
# y_ticks2 = np.arange(0,201,40)
# y_ticks3 = np.arange(0,20,2)
# xticks = np.arange(0,6,1)

# # Plot excluding annotation
# plt.figure(figsize=(7.5,6.5))
# lw = 0.7

# ax1 = plt.subplot(3,1,1)
# plt.plot(sqi_all,mpm_gpr_all,marker='o',linestyle=stylesgpr[0],color=colsgpr[0],label='GPR (+U95%)',linewidth=lw)
# plt.plot(sqi_all,mpm_mfgpr_all,marker='d',linestyle=stylesmfgpr[0],color=colsmfgpr[0],label='MF-GPR (+U95%)',linewidth=lw)
# plt.errorbar(sqi_all,mpm_gpr_all,yerr=mpmu95_gpr_all/2,color=colsgpr[0],capsize=3,linestyle='none')
# plt.errorbar(sqi_all,mpm_mfgpr_all,yerr=mpmu95_mfgpr_all/2,color=colsmfgpr[0],capsize=3,linestyle='none')
# plt.plot(0,MPM_conv,marker='o',linestyle='none',color='darkorange',label='Conventional (+U95%)')
# plt.errorbar(0,MPM_conv,yerr=MPMU95_conv/2,color='darkorange',capsize=3)
# plt.plot([-1,5.5],[MPM_true_MCS,MPM_true_MCS],linestyle=':',color='grey')    
# plt.text(0.26,MPM_true_MCS*0.975,'True MCS',color='grey')
# plt.legend(loc='upper right',ncol=3)
# plt.ylabel('$\widehat{C\'\'}$ [m]')
# plt.xticks(ticks=xticks,labels=[])
# plt.yticks(ticks=y_ticks1)
# plt.ylim([9,12])
# plt.xlim([-0.2,3.5])

# ax2 = plt.subplot(3,1,2)
# plt.plot(sqi_all,nev_gpr_all,marker='o',linestyle=stylesgpr[0],color=colsgpr[0],label='GPR',linewidth=lw)
# plt.plot(sqi_all,nev_mfgpr_all,marker='d',linestyle=stylesmfgpr[0],color=colsmfgpr[0],label='MF-GPR',linewidth=lw)
# plt.yticks(ticks=y_ticks2)
# plt.ylabel('# required HF samples')
# plt.xticks(ticks=xticks,labels=[])
# plt.ylim([0,200])
# plt.xlim([-0.2,3.5])
# plt.legend(loc='upper right',ncol=2)

# ax3 = plt.subplot(3,1,3)
# plt.fill_between(sqi_all,time_gpr_low,time_gpr_hig,facecolor=colsgpr[0], alpha=0.15,label='GPR (estimate)')
# plt.fill_between(sqi_all,time_mfgpr_low,time_mfgpr_hig,facecolor=colsmfgpr[0], alpha=0.15,label='MF-GPR (estimate)')
# plt.plot(0,HFtime_conv,marker='o',linestyle='none',color='darkorange',label='Conventional')
# plt.yticks(ticks=y_ticks3)
# plt.xlabel('SQI [-]') 
# plt.ylabel('Estimated time to \n simulate with HF tool [h]')
# plt.xticks(ticks=xticks)
# plt.ylim([0,8])
# plt.xlim([-0.2,3.5])
# plt.legend(loc='upper right',ncol=3)

# plt.savefig(f'{out_figs}/FigXX_PredVsSQI_allkernsMat32_HF-{sampopt_txt}_MF-{sampopt_txt}_HFacq-{acqname}_MFacq-{acqname}_exclann.pdf',bbox_inches='tight')


# # Alternative x-axis
# plt.figure(figsize=(6.5,4))
# x = np.arange(3)
# lw = 0.7

# plt.subplot(3,1,1)
# plt.plot(x,np.array([MPM_conv,np.nan,np.nan]),marker='o',linestyle='none',color='darkorange',label='Conventional')
# plt.errorbar(x,np.array([MPM_conv,np.nan,np.nan]),yerr=np.array([MPMU95_conv/2,np.nan,np.nan]),color='darkorange',capsize=3)
# plt.plot(x,np.append(np.nan,mpm_gpr_all),marker='o',linestyle=stylesgpr[0],color=colsgpr[0],label='GPR',linewidth=lw)
# plt.plot(x,np.append(np.nan,mpm_mfgpr_all),marker='d',linestyle=stylesmfgpr[0],color=colsmfgpr[0],label='MF-GPR',linewidth=lw)
# plt.errorbar(x,np.append(np.nan,mpm_gpr_all),yerr=np.append(np.nan,mpmu95_gpr_all/2),color=colsgpr[0],capsize=3,linestyle='none')
# plt.errorbar(x,np.append(np.nan,mpm_mfgpr_all),yerr=np.append(np.nan,mpmu95_mfgpr_all/2),color=colsmfgpr[0],capsize=3,linestyle='none')
# plt.plot([-0.2,2.2],[MPM_true_MCS,MPM_true_MCS],linestyle=':',color='grey')    
# plt.text(0.26,MPM_true_MCS*0.95,'True MCS',color='grey')
# plt.fill_between(np.array([-0.2,2.2]),(MPM_true_MCS-0.5*MPM_true_MCS_50hseeds_U95)*np.array([1,1]),(MPM_true_MCS+0.5*MPM_true_MCS_50hseeds_U95)*np.array([1,1]),facecolor='grey', alpha=0.3)
# plt.legend(loc='upper left',ncol=3)
# plt.ylabel('$\widehat{C\'\'}$ [m]')
# plt.xticks(ticks=xticks,labels=[])
# plt.yticks(ticks=y_ticks1)
# plt.ylim([9.5,12])
# plt.xlim(-0.3,2.3)

# ax2 = plt.subplot(3,1,2)
# plt.plot(x,np.append(np.nan,nev_gpr_all),marker='o',linestyle=stylesgpr[0],color=colsgpr[0],label='GPR',linewidth=lw)
# plt.plot(x,np.append(np.nan,nev_mfgpr_all),marker='d',linestyle=stylesmfgpr[0],color=colsmfgpr[0],label='MF-GPR',linewidth=lw)
# plt.yticks(ticks=y_ticks2)
# plt.ylabel('# HF samples\nat convergence')
# plt.xticks(ticks=xticks,labels=[])
# plt.ylim([0,80])
# plt.xlim(-0.3,2.3)

# plt.subplot(3,1,3)     
# plt.plot(x,np.array([HFtime_conv,np.nan,np.nan]),marker='o',linestyle='none',color='darkorange',label='Conventional')
# plt.fill_between(x,np.append(np.nan,time_gpr_low),np.append(np.nan,time_gpr_hig),facecolor=colsgpr[0], alpha=0.15,label='GPR (estimate)')
# plt.fill_between(x,np.append(np.nan,time_mfgpr_low),np.append(np.nan,time_mfgpr_hig),facecolor=colsmfgpr[0], alpha=0.15,label='MF-GPR (estimate)')
# plt.yticks(ticks=y_ticks3)
# plt.ylabel('Time to simulate\nwith HF tool [h]')
# plt.xticks(ticks=x,labels=['','GoodInd\nSQI=1.64','WorseInd\nSQI=3.07'])
# plt.ylim([0,6])
# # plt.legend(loc='best',ncol=3)      
# plt.xlim(-0.3,2.3)
    
# plt.savefig(f'{out_figs}/FigXX_PredVsSQI_allkernsMat32_HF-{sampopt_txt}_MF-{sampopt_txt}_HFacq-{acqname}_MFacq-{acqname}_new.pdf',bbox_inches='tight')
