# -*- coding: utf-8 -*-
"""
Created on Thu Aug 10 16:25:13 2023

@author: smvanessen

"""

# Clear all variables
from IPython import get_ipython
get_ipython().magic('reset -sf')

# Standard toolboxes
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import glob
import os
import re

# Add some other functions
from functions.subsample_routines import subsample_binned
from functions.distribution_scripts import MPM_from_DNR
from functions.theoretical_distributions import make_cdf_Forristall as Forr
from functions.T_enc import calc_T_enc

# Convergence criterion
from functions.convergenceAS import stopping_crit_AS

# Generalised Pareto distribution fitting (for true experimental HF reference)
from functions.GP_functions import GP_pexc,GP_pexc_log
from scipy.stats import genpareto
from scipy.optimize import curve_fit


# Close all figures
plt.close('all')


#############
# START INPUT
#############

# Plotsize
fsize_h = 7
fsize_v = 4.5

# Number of HF samples to include in plot
sampmaxplot1 = 70.5 # Convergence
sampstep1 = 10
sampmaxplot2 = 60.5 # All considered samples
sampstep2 = 10

# Acquisition function type to evaluate.
acqnames = ['USMV']
# acqnames = ['MD']

# Select used initial sampling option. Note: you can only select options for 
# which you first generated data with the Adaptive Screening method. For a 
# description of options A, B and C, also see Adaptive Screening. 
# Options: 'A', 'B' or 'C'
sampopt = 'A'

# Folder where the input CSV files are (generated by the Adaptive Screening procedure) (in)
# Folder where we write the output figures (out)
in_files = 'out_files_gw'
out_figs = 'out_figs_paper'

# Corresponding line styles and colors in the plots
stylesgpr = ['-.'];     colsgpr = ['green']  # matching distribution plots
stylesmfgpr = ['--'];   colsmfgpr = ['blue'] # matching distribution plots

# Kernel name single-fidelity HF-GPR (preferred Mat32 for all)
HFkern = 'Mat32'
MFkern1 = 'Mat32' 
MFkern2 = 'Mat32'

# Spectral properties basis wave system (for ref. Forristall and Pexp)
dur = 300 # target duration [s] 5 min
Tp = 9.7 # peak wave period [s]
Vs = 4.6*0.5144444 # current speed or vessel speed [m/s]
mu = np.pi # heading [rad] (180 deg = pi rad = head waves)

# Annotation switch for the plot of the MPM as function of iterations until 
# convergence. Paper figure 12 includes annotation.
# 1 = with annotation, 0 = without annotation
annon = 1

# HF response symbol, x-axis name and unit
# HFsymb = 'F_x'
HFsym = 'F_x \prime\prime'
HFxas = 'f'
HFunit = 'kN'

# Convergence limits stopping criterion (NOT USED; ONLY 10 SAMPLES PLOTTED)
limMAE = 5e6 # [Nm]
limCOV = 0.003


#############
# END INPUT
#############

# Check wether output directory exist, if not make it
isExist1 = os.path.exists(out_figs)
if not isExist1:
   os.makedirs(out_figs)

# Initialise
plt.figure(10000,figsize=(12.5,3))
plt.figure(20000,figsize=(12.5,3))
plt.figure(30000,figsize=(12.5,3))
x_ticks = np.arange(0,202,25)

# Initial sample options (see Adaptive Screening)
if sampopt == 'A':
    sampopt_txt = 'A_LFsubsamp'; N_hf = 2
    txtadd = f'A_LFsubsamp_exp{np.int(dur/60)}min'
elif sampopt == 'B':
    sampopt_txt = 'B_LFfit'; N_hf = 2
    txtadd = f'B_LFfit_exp{np.int(dur/60)}min'
elif sampopt == 'C':
    sampopt_txt = 'C_LFisHFfit'; N_hf = 2
    txtadd = f'C_LFisHFfit_exp{np.int(dur/60)}min'

# Derive target probability of exceedance, using the average number of wave encounters in target duration
Tpe = calc_T_enc(Tp,Vs,mu)
ne_exp = dur/Tpe
pexp = 1/ne_exp

# Initialise
conv_mpm_gpr = np.nan*np.zeros((4,len(acqnames)))
conv_mpmu95_gpr = np.nan*np.zeros((4,len(acqnames)))
conv_mpm_mfgpr = np.nan*np.zeros((4,len(acqnames)))
conv_mpmu95_mfgpr = np.nan*np.zeros((4,len(acqnames)))


# Loop over indicators
for nn in range(2):
    
    # Noise texts (figure titles and names)
    if nn == 0:
        indcase = 'PeaksFXCoarse_MAR3'
        # indalias = 'Coarse mesh ComFLOW green water force peaks'
        indalias = 'F_x \prime'
        # LFsym = 'F_{x,c}'
        LFsym = 'F_x \prime'
        LFxas = 'f'
        LFunit = 'kN'
    elif nn == 1:
        indcase = 'PeaksRelSBA_MAR3'
        # indalias = 'Coarse mesh ComFLOW RWE peaks at fore deck'
        indalias = 'R_{deck} \prime'
        LFsym = 'R_{deck} \prime'
        LFxas = 'r'
        LFunit = 'm'

    # Read experimental LF and HF true and training time traces from CSV file   
    DatLFtrue = np.loadtxt(f'{in_files}/TrueLFdata_GW_{indcase}_{txtadd}.csv', delimiter=',')
    v_true_l = DatLFtrue[:,0]
    p_true_l = DatLFtrue[:,1]
    DatHFtrue = np.loadtxt(f'{in_files}/TrueHFdata_GW_{indcase}_{txtadd}.csv', delimiter=',')
    v_true_h = DatHFtrue[:,0]
    p_true_h = DatHFtrue[:,1]
    DatLFtrain = np.loadtxt(f'{in_files}/TrainLFdata_GW_{indcase}_{txtadd}.csv', delimiter=',')
    v_train_l = DatLFtrain[:,0]
    p_train_l = DatLFtrain[:,1]
    
    # # Subsample the true distribution data for plotting later
    # # binsize = 0.05
    # binsize = 0.02
    # bins = np.arange(0,15,binsize)
    # v_true_l_sampled,p_true_l_sampled = subsample_binned(v_true_l,p_true_l,bins)
    # v_true_h_sampled,p_true_h_sampled = subsample_binned(v_true_h,p_true_h,bins)
    
    # Max values (for plotting) 
    minVal = np.floor(0.9*np.nanmin(v_true_h)) # HF
    maxVal = np.ceil(1.1*np.nanmax(v_true_h)) # HF
    maxVal_lf = np.ceil(1.1*np.nanmax(v_true_l)) # LF
    
    # Min PoE value (for plotting)
    minP = 0.9*np.nanmin(p_true_h) # HF
    
    
    
    ##############
    # Different w.r.t. wave only case: HF true data is fitted with Generalised 
    # Pareto to obtain true experimental reference. This is necessary because we
    # have fewer 'true' HF points than for the wave only case.
    ##############
    
    # Percentage of samples to consider
    perc = 10
    
    # Define function
    function_gp = GP_pexc_log
    
    # Apply to log of PoE, for better fitting accuracy
    p_true_h_log = np.log(p_true_h)
    
    # Define data to fit the function to 
    i_perc = int(len(v_true_h)*perc/100)
    xdata = np.squeeze(v_true_h[-19:])
    ydata_log = np.squeeze(p_true_h_log[-19:])
    
    # Perform curve fitting
    parameters, covariance_fit = curve_fit(function_gp, xdata, ydata_log)
    
    # Extract outcome
    v_true_h_fit = np.linspace(minVal*2,maxVal/1.5,100)
    loc = parameters[0]; scale = parameters[1];  shape = parameters[2]
    p_true_h_fit_log = function_gp(v_true_h_fit, loc, scale, shape)
    p_true_h_fit = np.exp(p_true_h_fit_log)
    
    # Calculate residuals (on linear scale, not log scale)
    residuals_fit = np.squeeze(np.exp(ydata_log)) - GP_pexc(xdata, *parameters)
    
    # Calculate / derive true MPM
    # MPM_true_EXP = MPM_from_DNR(v_true_h,p_true_h,ne_exp) # ORIGINAL BASED ON THE HF SAMPLES ONLY
    MPM_true_EXP = MPM_from_DNR(v_true_h_fit,p_true_h_fit,ne_exp) # MODIFIED USING THE FIT
    
    # Debugging: check of fit
    plt.figure()
    plt.yscale('log')
    plt.plot(v_true_h,p_true_h,'.b',label='HF experimental samples')
    plt.plot(v_true_h_fit,p_true_h_fit,'--r',label=f'GP fit to highest {perc}% HF experimental samples')
    plt.legend()
    
    ##############
    # End of different part
    ##############


        
    # Loop over selected acquisition functions (only functions that are available for both GPR and MF-GPR)
    for mm in range(len(acqnames)):
        acqname = acqnames[mm]
        print(f'Case {indalias} ({nn}/2), acq. function {acqname} ({mm}/{len(acqnames)-1})')
    
        #### Find last available iteration (in this case probably #200, as we later interpolate in this database using a convergence criterion)
        
        # Load data from HF-GPR and MF-GPR with all iteration results (MPM etc.)
        file_conv_hf = f'{in_files}/Conv_HF-GPR_GW_{indcase}_Set-{HFkern}_{txtadd}_{acqname}_Nhf{N_hf}.csv'
        file_conv_mf = f'{in_files}/Conv_MF-GPR_GW_{indcase}_Set-{MFkern1}-{MFkern2}_{txtadd}_{acqname}_Nhf{N_hf}.csv'
        
        dat_conv_hf_all = pd.read_csv(file_conv_hf, delimiter=',')
        dat_conv_mf_all = pd.read_csv(file_conv_mf, delimiter=',')
        
        # Extract last iteration number and max. number of samples
        nosamp_hf = dat_conv_hf_all['NoSamp'].to_numpy()
        nosamp_mf = dat_conv_mf_all['NoSamp'].to_numpy()
        
        # Max. number of samples, and max. number of iterations
        maxsamp_hf = int(np.nanmax(nosamp_hf))
        maxsamp_mf = int(np.nanmax(nosamp_mf))
        maxit_hf = int(maxsamp_hf-N_hf)
        maxit_mf = int(maxsamp_mf-N_hf)
        
        print(f'{indalias} with HF-GPR and {acqname}: available are {maxit_hf} iterations ({maxsamp_hf} samples)')
        print(f'{indalias} with MF-GPR and {acqname}: available are {maxit_mf} iterations ({maxsamp_mf} samples)')
        
        
        #### Load data for this last iteration and plot convergence
       
        # Number of iterations (not samples!)
        # noit_GPR = nosamp_hf-N_hf
        # noit_MFGPR = nosamp_mf-N_hf
    
        # Actual MPM values
        mpms_GPR = dat_conv_hf_all['MPM_HFGPR'].to_numpy()
        mpmsU95_GPR = dat_conv_hf_all['U95_MPM_HFGPR'].to_numpy()
        
        # mpms_MFGPR = dat_conv_mf_all['MPM_MFGPR'].to_numpy()
        # mpmsU95_MFGPR = dat_conv_mf_all['U95_MPM_MFGPR'].to_numpy()
        mpms_MFGPR = dat_conv_mf_all['MPM_MFGPR'].to_numpy() # typo in key, to fix in Adaptive Screening script, fix here afterwards
        mpmsU95_MFGPR = dat_conv_mf_all['U95_MPM_MFGPR'].to_numpy() # typo in key, to fix in Adaptive Screening script, fix here afterwards
        
        # Plot all iterations
        plt.figure(20000)
        plt.subplot(1,2,nn+1) 
        if nn == 0:
            # plt.ylabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (HFsymb,HFunit), fontsize=12)
            plt.ylabel(r'$\widehat{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (HFsym,HFunit), fontsize=12)
        plt.xlabel('Number of samples $j$ [-]')
        plt.xticks(ticks=x_ticks)
        plt.title(r'Indicator: $%s$' % indalias)
        plt.plot(nosamp_hf,mpms_GPR,linestyle=stylesgpr[mm],color=colsgpr[mm],label='GPR')  
        plt.plot(nosamp_mf,mpms_MFGPR,linestyle=stylesmfgpr[mm],color=colsmfgpr[mm],label='MF-GPR')    
        plt.fill_between(nosamp_hf.flatten(), (mpms_GPR - 0.5*mpmsU95_GPR).flatten(), 
                          (mpms_GPR + 0.5*mpmsU95_GPR).flatten(), facecolor=colsgpr[0], alpha=0.3,label='GPR U95%')
        plt.fill_between(nosamp_mf.flatten(), (mpms_MFGPR - 0.5*mpmsU95_MFGPR).flatten(), 
                          (mpms_MFGPR + 0.5*mpmsU95_MFGPR).flatten(), facecolor=colsmfgpr[0], alpha=0.3,label='MF-GPR U95%')
        plt.plot([0,sampmaxplot2],[MPM_true_EXP,MPM_true_EXP],linestyle=':',color='black'); plt.text(sampmaxplot2-30,MPM_true_EXP+0.03,'True experiments',color='black')
        plt.xlim([N_hf-1,sampmaxplot2])
        plt.ylim([0,10000])
        plt.xticks(ticks=np.arange(0,sampmaxplot2,sampstep2,dtype=int),label=[])
        if nn == 0:
            plt.legend(ncol=2)
        
        
        #### Apply stopping criterion to determine at which iteration we are converged
        
        # File name start of CSV with predicted data per iteration (exclude "itXX.csv", will be added later)
        fn_preds_GPR = f'{in_files}/HFGPR_HFpred_GW_{indcase}_Set-{HFkern}-nf_{txtadd}_HFacq-{acqname}_Nhf{N_hf}_HF'
        casename_GPR = f'{indcase} - GPR - {acqname}'
        min_it1 = 5
        ploton1 = False
        convit1,convsamp1 = stopping_crit_AS(nosamp_hf,mpms_GPR,mpmsU95_GPR,fn_preds_GPR,casename_GPR,pexp,min_it1,limMAE,limCOV,ploton1) # final
        
        fn_preds_MFGPR = f'{in_files}/MFGPR_HFpred_GW_{indcase}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_MFacq-{acqname}_Nhf{N_hf}_MF'
        casename_MFGPR = f'{indcase} - MF-GPR - {acqname}'
        min_it2 = 5
        ploton2 = False
        convit2,convsamp2 = stopping_crit_AS(nosamp_mf,mpms_MFGPR,mpmsU95_MFGPR,fn_preds_MFGPR,casename_MFGPR,pexp,min_it2,limMAE,limCOV,ploton2) # final

        # Index of 10 samples
        i_samp10_hf = 10-N_hf
        
        # Fraction step for help lines (-2*stp,-stp,0,stp,2*stp)
        stp = 0.2
        
        # Prepare for annotation
        if annon == 1:
            gpr10 = np.round((mpms_GPR[i_samp10_hf]/MPM_true_EXP-1)*100,1);  signgpr10 = ''
            if gpr10>0: 
                signgpr10 = '+' 
            mfgpr10 = np.round((mpms_MFGPR[i_samp10_hf]/MPM_true_EXP-1)*100,1); signmfgpr10 = ''
            if mfgpr10>0: 
                signmfgpr10 = '+' 
            if np.isfinite(convit1):
                gprconv = np.round((mpms_GPR[convit1]/MPM_true_EXP-1)*100,1); signgprconv = ''
                if gprconv>0: 
                    signgprconv = '+' 
            if np.isfinite(convit2):
                mfgprconv = np.round((mpms_MFGPR[convit2]/MPM_true_EXP-1)*100,1); signmfgprconv = ''
                if mfgprconv>0: 
                    signmfgprconv = '+'  
            
        # Same but all in one plot
        plt.figure(10000)
        plt.subplot(1,2,nn+1) 
        plt.plot([0,sampmaxplot1],(1-stp)*np.array([MPM_true_EXP,MPM_true_EXP]),linestyle=':',color='lightgrey',linewidth=0.3)
        plt.plot([0,sampmaxplot1],(1+stp)*np.array([MPM_true_EXP,MPM_true_EXP]),linestyle=':',color='lightgrey',linewidth=0.3)
        plt.plot([0,sampmaxplot1],(1-2*stp)*np.array([MPM_true_EXP,MPM_true_EXP]),linestyle=':',color='lightgrey',linewidth=0.3)
        plt.plot([0,sampmaxplot1],(1+2*stp)*np.array([MPM_true_EXP,MPM_true_EXP]),linestyle=':',color='lightgrey',linewidth=0.3)
        if nn == 0:
            # plt.ylabel('$\widehat{F_x\'\'}$ [kN]')
            plt.ylabel(r'$\widehat{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (HFsym,HFunit), fontsize=12)
            plt.text(0.93*sampmaxplot1,(1-stp)*MPM_true_EXP+0.03,f'-{int(stp*100)}%',color='lightgrey',fontsize=8)
            plt.text(0.93*sampmaxplot1,(1+stp)*MPM_true_EXP+0.03,f'+{int(stp*100)}%',color='lightgrey',fontsize=8)
            plt.text(0.93*sampmaxplot1,(1-2*stp)*MPM_true_EXP+0.03,f'-{int(2*stp*100)}%',color='lightgrey',fontsize=8)
            plt.text(0.93*sampmaxplot1,(1+2*stp)*MPM_true_EXP+0.03,f'+{int(2*stp*100)}%',color='lightgrey',fontsize=8)
        plt.xlabel('Number of samples $j$ [-]')
        plt.xticks(ticks=x_ticks)
        plt.title(r'Indicator: $%s$' % indalias)
        if np.isfinite(convsamp1):
            plt.plot(nosamp_hf[:convsamp1-1],mpms_GPR[:convsamp1-1],linestyle=stylesgpr[mm],color=colsgpr[mm],label='GPR')
        else:
            plt.plot(nosamp_hf,mpms_GPR,linestyle=stylesgpr[mm],color=colsgpr[mm],label='GPR')
        if np.isfinite(convsamp2):
            plt.plot(nosamp_mf[:convsamp2-1],mpms_MFGPR[:convsamp2-1],linestyle=stylesmfgpr[mm],color=colsmfgpr[mm],label='MF-GPR')  
        else:
            plt.plot(nosamp_mf,mpms_MFGPR,linestyle=stylesmfgpr[mm],color=colsmfgpr[mm],label='MF-GPR')  
        if np.isfinite(convsamp1):
            plt.fill_between(nosamp_hf[:convsamp1-1].flatten(), (mpms_GPR[:convsamp1-1] - 0.5*mpmsU95_GPR[:convsamp1-1]).flatten(), 
                              (mpms_GPR[:convsamp1-1] + 0.5*mpmsU95_GPR[:convsamp1-1]).flatten(), facecolor=colsgpr[0], alpha=0.3,label='GPR U95%')
        else:
            plt.fill_between(nosamp_hf.flatten(), (mpms_GPR - 0.5*mpmsU95_GPR).flatten(), 
                              (mpms_GPR + 0.5*mpmsU95_GPR).flatten(), facecolor=colsgpr[0], alpha=0.3,label='GPR U95%')
        if np.isfinite(convsamp2):
            plt.fill_between(nosamp_mf[:convsamp2-1].flatten(), (mpms_MFGPR[:convsamp2-1] - 0.5*mpmsU95_MFGPR[:convsamp2-1]).flatten(), 
                              (mpms_MFGPR[:convsamp2-1] + 0.5*mpmsU95_MFGPR[:convsamp2-1]).flatten(), facecolor=colsmfgpr[0], alpha=0.3,label='MF-GPR U95%')
        else:
            plt.fill_between(nosamp_mf.flatten(), (mpms_MFGPR - 0.5*mpmsU95_MFGPR).flatten(), 
                              (mpms_MFGPR + 0.5*mpmsU95_MFGPR).flatten(), facecolor=colsmfgpr[0], alpha=0.3,label='MF-GPR U95%')      
        plt.plot([0,sampmaxplot1],[MPM_true_EXP,MPM_true_EXP],linestyle=':',color='black'); plt.text(0.85*sampmaxplot1,MPM_true_EXP+0.03,'True exp.',color='black')
        plt.ylim([0,10000])
        plt.xlim([N_hf-1,sampmaxplot1])
        plt.xticks(ticks=np.arange(0,sampmaxplot1,sampstep1,dtype=int),label=[])
        if nn == 0:
            plt.legend(ncol=2)
        if annon == 1:
            plt.plot([nosamp_hf[i_samp10_hf],nosamp_hf[i_samp10_hf]],[MPM_true_EXP,9000],color='grey',linewidth=0.2)
            plt.text(nosamp_hf[i_samp10_hf]+0.1,7000,f'10 samples\nGPR: {signgpr10}{gpr10}%\nMF-GPR: {signmfgpr10}{mfgpr10}%',color='grey',fontsize=8)
            if np.isfinite(convsamp1):
                plt.plot([nosamp_hf[convsamp1-1]-0.1,nosamp_hf[convsamp1-1]-1],[mpms_GPR[convsamp1-1],11050],color='green',linewidth=0.2)
                plt.text(nosamp_hf[convsamp1-1],10900,f'converged\nGPR: {signgprconv}{gprconv}%',color='green',fontsize=8)
            if np.isfinite(convsamp2):
                plt.plot([nosamp_mf[convsamp2-1]-0.1,nosamp_mf[convsamp2-1]-1],[mpms_MFGPR[convsamp2-1],9250],color='blue',linewidth=0.2)
                plt.text(nosamp_mf[convsamp2-1],9200,f'converged\nMF-GPR: {signmfgprconv}{mfgprconv}%',color='blue',fontsize=8)
        
        
        # Same but all in one plot NOW ONLY MF-GPR
        plt.figure(30000)
        plt.subplot(1,2,nn+1) 
        plt.plot([0,sampmaxplot1],(1-stp)*np.array([MPM_true_EXP,MPM_true_EXP]),linestyle=':',color='lightgrey',linewidth=0.3)
        plt.plot([0,sampmaxplot1],(1+stp)*np.array([MPM_true_EXP,MPM_true_EXP]),linestyle=':',color='lightgrey',linewidth=0.3)
        plt.plot([0,sampmaxplot1],(1-2*stp)*np.array([MPM_true_EXP,MPM_true_EXP]),linestyle=':',color='lightgrey',linewidth=0.3)
        plt.plot([0,sampmaxplot1],(1+2*stp)*np.array([MPM_true_EXP,MPM_true_EXP]),linestyle=':',color='lightgrey',linewidth=0.3)
        if nn == 0:
            # plt.ylabel('$\widehat{F_x\'\'}$ [kN]')
            plt.ylabel(r'$\widehat{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (HFsym,HFunit), fontsize=12)
            plt.text(0.93*sampmaxplot1,(1-stp)*MPM_true_EXP+0.03,f'-{int(stp*100)}%',color='lightgrey',fontsize=8)
            plt.text(0.93*sampmaxplot1,(1+stp)*MPM_true_EXP+0.03,f'+{int(stp*100)}%',color='lightgrey',fontsize=8)
            plt.text(0.93*sampmaxplot1,(1-2*stp)*MPM_true_EXP+0.03,f'-{int(2*stp*100)}%',color='lightgrey',fontsize=8)
            plt.text(0.93*sampmaxplot1,(1+2*stp)*MPM_true_EXP+0.03,f'+{int(2*stp*100)}%',color='lightgrey',fontsize=8)
        plt.xlabel('Number of samples $j$ [-]')
        plt.xticks(ticks=x_ticks)
        plt.title(r'Indicator: $%s$' % indalias)
        if np.isfinite(convit2):
            plt.plot(nosamp_mf[:convit2],mpms_MFGPR[:convit2],linestyle=stylesmfgpr[mm],color=colsmfgpr[mm],label='MF-GPR')  
        else:
            plt.plot(nosamp_mf,mpms_MFGPR,linestyle=stylesmfgpr[mm],color=colsmfgpr[mm],label='MF-GPR')  
        if np.isfinite(convit2):
            plt.fill_between(nosamp_mf[:convit2].flatten(), (mpms_MFGPR[:convit2] - 0.5*mpmsU95_MFGPR[:convit2]).flatten(), 
                              (mpms_MFGPR[:convit2] + 0.5*mpmsU95_MFGPR[:convit2]).flatten(), facecolor=colsmfgpr[0], alpha=0.3,label='MF-GPR U95%')
        else:
            plt.fill_between(nosamp_mf.flatten(), (mpms_MFGPR - 0.5*mpmsU95_MFGPR).flatten(), 
                              (mpms_MFGPR + 0.5*mpmsU95_MFGPR).flatten(), facecolor=colsmfgpr[0], alpha=0.3,label='MF-GPR U95%')      
        plt.plot([0,sampmaxplot1],[MPM_true_EXP,MPM_true_EXP],linestyle=':',color='black'); plt.text(0.85*sampmaxplot1,MPM_true_EXP+0.03,'True exp.',color='black')
        plt.ylim([0,8000])
        plt.xlim([N_hf-1,sampmaxplot1])
        plt.xticks(ticks=np.arange(0,sampmaxplot1,sampstep1,dtype=int),label=[])
        if nn == 0:
            plt.legend(ncol=2)
        if annon == 1:
            plt.plot([nosamp_hf[i_samp10_hf],nosamp_hf[i_samp10_hf]],[MPM_true_EXP,9000],color='grey',linewidth=0.2)
            plt.text(nosamp_hf[i_samp10_hf]+0.1,7000,f'10 samples\nMF-GPR: {signmfgpr10}{mfgpr10}%',color='grey',fontsize=8)
            if np.isfinite(convit2):
                plt.plot([nosamp_mf[convit2]-0.1,nosamp_mf[convit2]-1],[mpms_MFGPR[convit2],9250],color='blue',linewidth=0.2)
                plt.text(nosamp_mf[convit2],9200,f'converged\nMF-GPR: {signmfgprconv}{mfgprconv}%',color='blue',fontsize=8)
            
        #### CALCULATE AND PRINT MPM VALUES AND U95% FOR CONVERGED RESULTS
        print(f'Case: {indalias}, {dur/3600}h:')
        print(f'MPM true experiments: {MPM_true_EXP} m')
        if np.isfinite(convit1):
            conv_mpm_gpr[nn,mm] = mpms_GPR[convit1]
            conv_mpmu95_gpr[nn,mm] = mpmsU95_GPR[convit1]
            conv_mpm_mfgpr[nn,mm] = mpms_MFGPR[convit2]
            conv_mpmu95_mfgpr[nn,mm] = mpmsU95_MFGPR[convit2]
            print('Converged results:')
        else:
            conv_mpm_gpr[nn,mm] = mpms_GPR[i_samp10_hf]
            conv_mpmu95_gpr[nn,mm] = mpmsU95_GPR[i_samp10_hf]
            conv_mpm_mfgpr[nn,mm] = mpms_MFGPR[i_samp10_hf]
            conv_mpmu95_mfgpr[nn,mm] = mpmsU95_MFGPR[i_samp10_hf]
            print('Results with 10 HF samples:')
        print(f'MPM GPR {acqname}: {conv_mpm_gpr[nn,mm]} m (w.r.t. EXP {(conv_mpm_gpr[nn,mm]/MPM_true_EXP-1)*100}\%), U95 {conv_mpmu95_gpr[nn,mm]} m')
        print(f'MPM MF-GPR {acqname}: {conv_mpm_mfgpr[nn,mm]} m (w.r.t. EXP {(conv_mpm_mfgpr[nn,mm]/MPM_true_EXP-1)*100}\%), U95 {conv_mpmu95_mfgpr[nn,mm]} m')
                

        ############### READ AND PLOT CONVERGED GPR RESULTS FOR PRESENT CASE & ACQ FUNCT
        
        # Maximum iteration number
        if ~np.isfinite(convit1):
            usedit1 = maxit_hf
        else:
            usedit1 = convit1   
        
        # GPR (2x)
        DatHFtrainIt1 = np.genfromtxt(f'{in_files}/HFGPR_TrainData_GW_{indcase}_Set-{HFkern}-nf_{txtadd}_HFacq-{acqname}_Nhf{N_hf}_HFit{usedit1}.csv', skip_header=1, delimiter=',')
        v_train_h1 = DatHFtrainIt1[:,0]
        p_train_h1 = DatHFtrainIt1[:,1]
        
        DatHFpredIt1 = np.genfromtxt(f'{in_files}/HFGPR_HFpred_GW_{indcase}_Set-{HFkern}-nf_{txtadd}_HFacq-{acqname}_Nhf{N_hf}_HFit{usedit1}.csv', skip_header=1, delimiter=',')
        Prange_predHF = DatHFpredIt1[:,0]
        mean_high_gp_model = DatHFpredIt1[:,1]
        var_high_gp_model = DatHFpredIt1[:,2]
        U95_high_gp_model = 1.96*np.sqrt(var_high_gp_model)
        
        ###### Plot HF
                
        # Plot the posterior mean and variance (U95) from the reference HF gaussian process
        plt.figure(1000+nn,figsize=(fsize_h,fsize_v))
        plt.yscale('log')
        # plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" exp.')
        plt.plot(v_true_h,p_true_h,marker='.',color='pink',linestyle='None',label='HF "true" exp.')
        plt.plot(v_true_h_fit,p_true_h_fit,color='pink',linestyle='--',label='GP fit to HF "true" exp. (validation)')
        plt.plot(v_train_h1,p_train_h1, color='r', marker='d',linestyle='None', label=f'{usedit1} HF samples')
        plt.plot(mean_high_gp_model,Prange_predHF,'-.',color='g',label='GPR prediction')
        plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_gp_model - 0.5*U95_high_gp_model).flatten(), 
                          (mean_high_gp_model + 0.5*U95_high_gp_model).flatten(), facecolor='g', alpha=0.1,label='GPR prediction U95%')
        plt.plot([0,maxVal],[pexp,pexp],':',color='black')
        plt.text(0.9*maxVal,pexp*1.4,'$P_{exp}$')
        plt.ylabel(r'$P(%s \geq %s)$ $\mathrm{[-]}$' % (HFsym,HFxas), fontsize=12)
        plt.xlabel(r'$%s$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (HFxas,HFunit), fontsize=12)
        plt.ylim([minP,1])
        plt.xlim([0,maxVal])
        plt.legend()
        plt.savefig(f'{out_figs}/FigXX_Prediction_HFGPR_{indcase}_Set-{HFkern}_{sampopt_txt}_{acqname}_it{usedit1}.pdf',bbox_inches='tight')
       
        
        
        ############### READ AND PLOT CONVERGED MF-GPR RESULTS FOR PRESENT CASE & ACQ FUNCT
        
        # Maximum iteration number
        if ~np.isfinite(convit2):
            usedit2 = maxit_mf
        else:
            usedit2 = convit2
        
        # MF-GPR (4x)
        DatLFtrainIt2 = np.genfromtxt(f'{in_files}/MFGPR_LFTrainData_GW_{indcase}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_MFacq-{acqname}_Nhf{N_hf}_MFit{usedit2}.csv', skip_header=1, delimiter=',')
        v_train_l2 = DatLFtrainIt2[:,0]
        p_train_l2 = DatLFtrainIt2[:,1]
        
        DatHFtrainIt2 = np.genfromtxt(f'{in_files}/MFGPR_HFTrainData_GW_{indcase}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_MFacq-{acqname}_Nhf{N_hf}_MFit{usedit2}.csv', skip_header=1, delimiter=',')
        v_train_h2 = DatHFtrainIt2[:,0]
        p_train_h2 = DatHFtrainIt2[:,1]
                
        DatLFpredIt2 = np.genfromtxt(f'{in_files}/MFGPR_LFpred_GW_{indcase}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_MFacq-{acqname}_Nhf{N_hf}_MFit{usedit2}.csv', skip_header=1, delimiter=',')
        Prange_predLF = DatLFpredIt2[:,0]
        mean_low_mfgp_model = DatLFpredIt2[:,1]
        var_low_mfgp_model = DatLFpredIt2[:,2]
        U95_low_mfgp_model = 1.96*np.sqrt(var_low_mfgp_model)
        
        DatHFpredIt2 = np.genfromtxt(f'{in_files}/MFGPR_HFpred_GW_{indcase}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_MFacq-{acqname}_Nhf{N_hf}_MFit{usedit2}.csv', skip_header=1, delimiter=',')
        Prange_predHF = DatHFpredIt2[:,0]
        mean_high_mfgp_model = DatHFpredIt2[:,1]
        var_high_mfgp_model = DatHFpredIt2[:,2]
        U95_high_mfgp_model = 1.96*np.sqrt(var_high_mfgp_model)
        
        # Debugging
        plt.figure()
        plt.yscale('log')
        # plt.plot(v_true_l_sampled,p_true_l_sampled,marker='.',color=(0.7,0.7,0.7),linestyle='None',label='LF "true" exp.')
        plt.plot(v_true_l,p_true_l,marker='.',color=(0.7,0.7,0.7),linestyle='None',label='LF "true" exp.')
        plt.plot(v_train_l2,p_train_l2, color='grey', marker='o',linestyle='None', label='LF samples')
        
        ###### Plot MF
        
        # Plot the posterior mean and variance (U95) from the MF gaussian process
        plt.figure(2000+nn,figsize=(fsize_h,fsize_v))
        # Plot the posterior mean and variance (U95) from the MF gaussian process  (use secondary x-axis for training data LF and HF)
        plt.yscale('log')
        plt.ylabel(r'$\mathrm{P(}\mathrm{%s} \geq \mathrm{%s} \mathrm{)}$   and   $\mathrm{P(}\mathrm{%s} \geq \mathrm{%s} \mathrm{)}$' % (HFsym,HFxas,LFsym,LFxas), fontsize=12)
        # -- x-axis1 (HF data)
        aX1c = plt.subplot()
        # leg1 = aX1c.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" exp.')
        # plt.plot(v_true_h_fit,p_true_h_fit,color='pink',linestyle='--',label='GP fit to HF "true" exp. (validation)')
        leg1 = aX1c.plot(v_true_h,p_true_h,marker='.',color='pink',linestyle='None',label='HF "true" exp. (validation)')
        leg2 = aX1c.plot(v_train_h2,p_train_h2, color='r', marker='d',linestyle='None',markersize=4, label=f'{len(v_train_h2)} HF samples')
        leg3 = aX1c.plot(mean_high_mfgp_model,Prange_predHF, '--', color='g',label='MF-GPR pred. HF (+U95%)')
        leg4 = aX1c.fill_betweenx(Prange_predHF.flatten(), (mean_high_mfgp_model - 0.5*U95_high_mfgp_model).flatten(), 
                         (mean_high_mfgp_model + 0.5*U95_high_mfgp_model).flatten(), facecolor='g', alpha=0.1,label='MF-GPR pred. HF U95% (label not printed)')
        leg5 = aX1c.plot([0,maxVal],[pexp,pexp],':',color='black') #,label='$P_{exp}$')
        aX1c.text(14,pexp*1.4,'$P_{exp}$')
        aX1c.set_xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (HFsym,HFunit), fontsize=12,color='red')
        aX1c.set_xlim(left=0,right=maxVal)
        # -- x-axis2 (LF data)
        aX2c = aX1c.twiny()
        # leg6 = aX2c.plot(v_true_l_sampled,p_true_l_sampled,marker='.',color=(0.7,0.7,0.7),linestyle='None',label='LF "true" exp.')
        leg7 = aX2c.plot(v_train_l2,p_train_l2, color='grey', marker='o',linestyle='None',markersize=4, label='LF samples')
        leg8 = aX2c.plot(mean_low_mfgp_model,Prange_predLF, '--', color='b',alpha=0.4,label='MF-GPR pred. LF (+U95%)')
        leg9 = aX2c.fill_betweenx(Prange_predLF.flatten(), (mean_low_mfgp_model - 0.5*U95_low_mfgp_model).flatten(), 
                          (mean_low_mfgp_model + 0.5*U95_low_mfgp_model).flatten(), facecolor='b', alpha=0.1,label='MF-GPR pred. U95% (label not printed)')
        aX2c.set_xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (LFsym,LFunit), fontsize=12,color=(0.4,0.4,0.4))
        aX2c.set_xlim(left=0,right=maxVal_lf)
        # -- y-axis
        plt.ylim([minP,1])
        # leg = leg6+leg7+leg2+leg8+leg3+leg1
        leg = leg7+leg2+leg8+leg3+leg1
        labs = [l.get_label() for l in leg]
        aX1c.legend(leg,labs,loc='best')            
        plt.savefig(f'{out_figs}/FigXX_Prediction_MFGPR_{indcase}_Set-{MFkern1}-{MFkern2}_{sampopt_txt}_{acqname}_it{usedit2}_1.pdf',bbox_inches='tight')
        
        # # Debugging
        # plt.plot(v_true_l_sampled,p_true_l_sampled,marker='.',color=(0.7,0.7,0.7),linestyle='None',label='LF "true" exp.')
        # handles, labels = plt.gca().get_legend_handles_labels() # get handles and labels (in order to change legend order)
        # order = [0,7,1,2,3,4] #specify order of items in legend
        # plt.legend([handles[idx] for idx in order],[labels[idx] for idx in order]) #add legend to plot
        # plt.savefig(f'{out_figs}/Fig110_Prediction_MFGPR_{indalias}_Set-{MFkern1}-{MFkern2}_{sampopt_txt}_{acqname}_it{usedit2}_inclLFall.pdf',bbox_inches='tight')
        
        # Alternative side-by-side plot
        plt.figure(2010+nn,figsize=(fsize_h*1.7,fsize_v*0.8))
        # -- subplot 2 (HF data)
        plt.subplot(1,2,2)
        plt.yscale('log') 
        # plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" exp. (validation)')
        plt.plot(v_true_h,p_true_h,marker='.',color='pink',linestyle='None',label='HF "true" experiments')
        plt.plot(v_true_h_fit,p_true_h_fit,color='pink',linestyle='--',label=f'GP fit to HF "true" exp. (validation)')
        plt.plot(v_train_h2,p_train_h2, color='r', marker='d',linestyle='None',markersize=4, label=f'{usedit2} HF samples')
        plt.plot(mean_high_mfgp_model,Prange_predHF, '--', color='b',label='MF-GPR pred. HF')
        plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_mfgp_model - 0.5*U95_high_mfgp_model).flatten(), 
                         (mean_high_mfgp_model + 0.5*U95_high_mfgp_model).flatten(), facecolor='b', alpha=0.1,label='MF-GPR pred. HF U95%')
        plt.plot([0,maxVal],[pexp,pexp],':',color='black') #,label='$P_{exp}$')
        plt.text(0.9*maxVal,pexp*1.4,'$P_{exp}$')
        # plt.ylabel('P(F_{x,c}\'\'$\geq$c) [-]')
        # plt.xlabel('f [kN]')
        plt.ylabel(r'$P(%s \geq %s)$ $\mathrm{[-]}$' % (HFsym,HFxas), fontsize=12)
        plt.xlabel(r'$%s$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (HFxas,HFunit), fontsize=12)
        plt.xlim(left=0,right=maxVal)
        plt.ylim([minP,1])
        plt.legend(loc='best')
        # -- subplot1 (LF data)
        plt.subplot(1,2,1)
        plt.yscale('log') 
        # # plt.plot(v_true_l_sampled,p_true_l_sampled,marker='.',color=(0.7,0.7,0.7),linestyle='None',label='LF "true" exp.')
        # plt.plot(v_true_l,p_true_l,marker='.',color=(0.7,0.7,0.7),linestyle='None',label='LF "true" exp.')
        plt.plot(v_train_l2,p_train_l2, color='grey', marker='o',linestyle='None',markersize=4, label='LF samples')
        plt.plot(mean_low_mfgp_model,Prange_predLF, '--', color='b',alpha=0.4,label='MF-GPR pred. LF')
        plt.fill_betweenx(Prange_predLF.flatten(), (mean_low_mfgp_model - 0.5*U95_low_mfgp_model).flatten(), 
                          (mean_low_mfgp_model + 0.5*U95_low_mfgp_model).flatten(), facecolor='b', alpha=0.1,label='MF-GPR pred. LF U95%')
        # plt.ylabel('P(F_x\'$\geq$c) [-]')
        # plt.xlabel('f [m]')
        plt.ylabel(r'$P(%s \geq %s)$ $\mathrm{[-]}$' % (LFsym,LFxas), fontsize=12)
        plt.xlabel(r'$%s$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (LFxas,LFunit), fontsize=12)
        plt.xlim(left=0,right=maxVal_lf)
        # -- y-axis
        plt.ylim([minP,1])
        plt.legend(loc='best')
        plt.savefig(f'{out_figs}/Fig12_Prediction_MFGPR_{indcase}_Set-{MFkern1}-{MFkern2}_{sampopt_txt}_{acqname}_it{usedit2}_2.pdf',bbox_inches='tight')
                
          
        ############### PLOT GPR AND MF-GPR IN SAME PLOT
        
        # ALL IN ONE Plot the posterior mean and variance (U95) from the MF gaussian process (without training data, as these are different for HF-GPR and MF-GPR)
        plt.figure(3000+nn,figsize=(fsize_h,fsize_v))
        plt.yscale('log')
        # plt.plot(v_true_l_sampled,p_true_l_sampled,marker='.',color=(0.7,0.7,0.7),linestyle='None',label='LF "true" exp.')
        plt.plot(v_true_l,p_true_l,marker='.',color=(0.7,0.7,0.7),linestyle='None',label='LF "true" exp. samples')
        # plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" exp.') 
        plt.plot(v_true_h,p_true_h,marker='.',color='pink',linestyle='None',label='HF "true" exp.') 
        plt.plot(mean_low_mfgp_model,Prange_predLF, '--', color='b',label='MF-GPR predictions')
        plt.fill_betweenx(Prange_predLF.flatten(), (mean_low_mfgp_model - 0.5*U95_low_mfgp_model).flatten(), 
                          (mean_low_mfgp_model + 0.5*U95_low_mfgp_model).flatten(), facecolor='b', alpha=0.3,label='MF-GPR predictions U95%')
        plt.plot(mean_high_mfgp_model,Prange_predHF, '--', color='b',label='MF-GPR prediction HF  (label not printed)')
        plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_mfgp_model - 0.5*U95_high_mfgp_model).flatten(), 
                          (mean_high_mfgp_model + 0.5*U95_high_mfgp_model).flatten(), facecolor='b', alpha=0.3,label='MF-GPR prediction HF U95% (label not printed)')
        plt.plot(mean_high_gp_model,Prange_predHF,'-.',color='g',label='GPR prediction HF')
        plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_gp_model - 0.5*U95_high_gp_model).flatten(), 
                          (mean_high_gp_model + 0.5*U95_high_gp_model).flatten(), facecolor='g', alpha=0.1,label='GPR prediction HF U95%')
        plt.plot([0,maxVal],[pexp,pexp],':',color='black')
        plt.text(0.9*maxVal,pexp*1.4,'$P_{exp}$')
        plt.ylabel(r'$P(%s \geq %s)$ $\mathrm{[-]}$' % (HFsym,HFxas), fontsize=12)
        plt.xlabel(r'$%s$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (HFxas,HFunit), fontsize=12)
        plt.ylim([minP,1])
        plt.xlim([0,maxVal])
        # plt.title(f'Adaptive GRR and MF-GPR \n GPR converged at it.{convit1} ({convit1+N_hf} samples) and MF-GPR at it.{usedit2} ({usedit2+N_hf} samples)')
        handles, labels = plt.gca().get_legend_handles_labels() # get handles and labels (in order to change legend order)
        order = [0,1,2,3,6,7] #specify order of items in legend
        plt.legend([handles[idx] for idx in order],[labels[idx] for idx in order]) #add legend to plot
        plt.savefig(f'{out_figs}/FigXX_AllPred0_{indcase}_HFSet-{HFkern}_MFSet-{MFkern1}-{MFkern2}_HF-{sampopt_txt}_MF-{sampopt_txt}_HFacq-{acqname}_MFacq-{acqname}_HFit{convit1}_MFit{usedit2}.pdf',bbox_inches='tight')

       
# Finalise and save plot
plt.figure(10000)
if annon == 1:
    plt.savefig(f'{out_figs}/FigXX_Convergence_AllCases_GPR-{sampopt_txt}_MFGPR-{sampopt_txt}_{acqname}_converged_ann.pdf',bbox_inches='tight')
else:
    plt.savefig(f'{out_figs}/FigXX_Convergence_AllCases_GPR-{sampopt_txt}_MFGPR-{sampopt_txt}_{acqname}_converged.pdf',bbox_inches='tight')

plt.figure(30000)
if annon == 1:
    plt.savefig(f'{out_figs}/Fig11_Convergence_AllCases_MFGPR-{sampopt_txt}_{acqname}_converged_ann.pdf',bbox_inches='tight')
else:
    plt.savefig(f'{out_figs}/FigXX_Convergence_AllCases_MFGPR-{sampopt_txt}_{acqname}_converged.pdf',bbox_inches='tight')
            
            
            
    
    
