# -*- coding: utf-8 -*-
"""
Created on 19 Oct 2023

@author: smvanessen
"""

# Clear all variables
from IPython import get_ipython
get_ipython().magic('reset -sf')

# Standard toolboxes
import numpy as np
import matplotlib.pyplot as plt
import sys 
import os

# Required to load Matlab .mat files
import scipy.io

# Distributions (Rayleigh and Forristall), first add path to functions
sys.path.append('functions')
from functions.theoretical_distributions import make_cdf_Rayleigh as Rayl
from functions.theoretical_distributions import make_cdf_Forristall as Forr
from functions.distribution_scripts import calc_distribution,MPM_from_DNR
from functions.T_enc import calc_T_enc

# Allow for subsampling of large datasets for plotting & saving as PDF
from functions.subsample_routines import subsample_binned

# Close all figures
plt.close('all')


#############
# START INPUT
#############
           
# Figure sizes (units unclear, see matplotlib doc)
fsize_h = 5.5
fsize_v = 5
  
# Test number
testno = '10050401'

# HF response input
HFsigname = 'FX'; HFnewname = 'FX accommodation'; HFunit = 'kN'; HFsymb = 'fx'

# Indicator quality case input (to load the right case study dataset). 
indcase = 'FXcoarse';
# indcase = 'RelSBAcoarse'; 

if indcase == 'FXcoarse':
    # Only available for single run 10050401   
    LFsigname = 'Peaks FXtot (coarse)'
    LFnewname = 'PeaksFXCoarse'
    LFunit = 'kN'
    LFsymb = 'fxc' 
elif indcase == 'RelSBAcoarse':
    # Available for all full tests
    LFsigname = 'Peaks RelSBA'
    LFnewname = 'PeaksRelSBA'
    LFunit = 'm'
    LFsymb = 'rs' 

# Used screening tool (required to read correct dataseet)
# ComFLOW 3 of MARIN; availabe for single run 10050401
toolLF = 'MAR3'

# Folder name where input files are stored (files to read)
inputfolder = 'inputfiles_gw'

# Folder name in which to store figures (rel. to present folder)
figfolder = 'out_figs_paper'
      
# Target duration [s]
dur = 600

# Total available duration [s]
Texp = 1800

# Wave properties, speed and heading (for derivation wave encounter period and Pexp)
Tp = 9.7 # [s]
Vs = 4.6*0.51444444 # [m/s]
mu = 180 # [deg]

#############
# END INPUT
#############

# Check wether output directories exist, if not make them
isExist = os.path.exists(figfolder)
if not isExist:
   os.makedirs(figfolder)
  
# Derive target exposure duration probability level
Tpe = calc_T_enc(Tp,Vs,mu)
ne_exp = dur/Tpe
pexp = 1/ne_exp

# Assemble file name - the 'oneprobe' data seem to the the latest versions
if testno == '100504':
    filen = f'{inputfolder}\{testno}_{LFsigname}_MatchedPeaks_IND_{HFsigname}_oneprobe_inclCFD.mat'
elif testno == '10050401': 
    filen = f'{inputfolder}\{testno}_{LFsigname}_CFD_MatchedPeaks_IND_{HFsigname}.mat'  
else:
    filen = f'{inputfolder}\{testno}_{LFsigname}_MatchedPeaks_IND_{HFsigname}_oneprobe.mat'    
    
# Load mat file with scatter plot data - for a description of the signals see * below
mat1 = scipy.io.loadmat(filen)

# Read matched LF and HF peaks from file
t_peaks = mat1[f't_p{toolLF}']
peaks_lf_all = np.squeeze(mat1[f'p{toolLF}'])
peaks_hf_all_scr = np.squeeze(mat1[f'mFXcorr_{toolLF}'])

# Read HF true peaks from file
filen_true = f'{inputfolder}\{testno}_FXpeaks_versus_Impulses.mat'
mat2 = scipy.io.loadmat(filen_true) 
# mat2.keys()
peaks_hf_all_true = np.squeeze(mat2['pFX'])

print('Reading true HF peak and screening matched LF-HF peak MAT files done')

# Calculate true distributions for 'Monte-Carlo' dataset (all data, validation)
v_true_l,p_true_l_temp,_ = calc_distribution(peaks_lf_all)
v_true_h,p_true_h_temp,_ = calc_distribution(peaks_hf_all_true)

#####
# Correct for using one estimated number of waves as basis instead of the number 
# of data points. Because the LF and HF signals can be different variables, we 
# need a common basis for their probability of exceedance (PoE) distributions. 
# This is provided by  Nwaves; the largest LF and HF value both have PoE 
# 1/Nwaves in the output.

# Estimated number of wave encounters in full experimental duration (where 
# the full experimental duration is estimated based on the time of the last 
# indicator peak). 
Nwaves = Texp/dur*ne_exp

# Correct for basis Nwaves instead of the number of LF or HF peaks
N_true_l = len(v_true_l)
N_true_h = len(v_true_h)
p_true_l = p_true_l_temp*N_true_l/Nwaves
p_true_h = p_true_h_temp*N_true_h/Nwaves

#####

# Calculate / derive true MPM from DNR
MPM_true_MCS = MPM_from_DNR(v_true_h,p_true_h,ne_exp)

# Initialise plots 
fig1 = plt.figure(1,figsize=(fsize_h,fsize_v))
ax1 = fig1.add_subplot(1,1,1)
ax1.set_yscale('log')
l, b, w, h = ax1.get_position().bounds
ax1.set_position([l, b, w, h*0.95]) # make less high so 3 line title fits

# Define colors
colMCS_LF = 'black'
colMCS_HF = 'red'
colrealLF = 'grey'
colrealHF = 'pink'

# Max. plot value
maxvalall = np.nanmax((np.nanmax(v_true_l),np.nanmax(v_true_h)))

# Plot true MCS DNR and Pexp + finish and save plot
ax1.plot(v_true_l,p_true_l,marker='.',color=colMCS_LF,linestyle='None',label='LF "true" MCS') 
ax1.plot(v_true_h,p_true_h,marker='.',color=colMCS_HF,linestyle='None',label='HF "true" MCS')
ax1.plot([0,maxvalall*1.1],[pexp,pexp],':',color='black')
ax1.text(maxvalall*0.9,pexp*1.4,'$P_{exp}$')
if indcase == 'FXcoarse':
    plt.ylabel('P(FX $\geq$ fx) [-]')
    plt.xlabel('fx [Nm]')
elif indcase == 'RelSBAcoarse':
    plt.ylabel('P(FX $\geq$ fx) or P(R $\geq$ r) [-]')
    plt.xlabel('fx [kN] or r [m]')
plt.ylim([10**(-5),1])
plt.xlim([0,maxvalall*1.1])
plt.legend()
plt.savefig(f'{figfolder}/Fig15_MCSinput_{indcase}.pdf',bbox_inches='tight')

print(f'{dur/3600}h MPM True MCS: {MPM_true_MCS} Nm')

