# -*- coding: utf-8 -*-
"""
Created on Thu Nov  3 11:58:45 2023
@author: smvanessen
"""

# Import standard toolboxes
import numpy as np
import matplotlib.pyplot as plt
# import sys
import pandas as pd
import os
import csv

# # Import GPy and Emukit functions. Refer to this documentation:
# # https://github.com/SheffieldML/GPy
# # https://gpy.readthedocs.io/en/deploy/
# # https://emukit.github.io/
# # We use the linear MF-GPR model, but there is also a (customisable) non-linear model: 
# # https://github.com/EmuKit/emukit/blob/main/emukit/multi_fidelity/models/non_linear_multi_fidelity_model.py
# import GPy
# import emukit.multi_fidelity
# from emukit.model_wrappers.gpy_model_wrappers import GPyMultiOutputWrapper, GPyModelWrapper
# from emukit.multi_fidelity.models import GPyLinearMultiFidelityModel           
# from emukit.multi_fidelity.convert_lists_to_array import convert_x_list_to_array, convert_xy_lists_to_arrays

# # Additions for Experimental Design / Adaptive Sampling loop (a.o. used for acquisition function MaxV)
# from emukit.core import ParameterSpace, ContinuousParameter

# # Acquisition functions from Emukit library
# from emukit.experimental_design.acquisitions import IntegratedVarianceReduction, ModelVariance

# Import custom acquisition functions as new classes
# sys.path.append('functions')
from functions.acq_maxval import MaxV
from functions.acq_parab_pdist import ParabPdist
from functions.acq_maxdist import MaxDist

# Import some other custom functions
from functions.distribution_scripts import calc_distribution,MPM_from_DNR,fit_weibull3_lsq
from functions.subsample_routines import subsample_binned,subsample_binned_2D
from functions.GP_functions import GP_pexc,GP_pexc_log

# # Filters
# from scipy.signal import savgol_filter
# from scipy.ndimage import gaussian_filter1d, uniform_filter1d
# from scipy.signal import medfilt

# Generalised Pareto distribution fitting
from scipy.stats import genpareto
from scipy.optimize import curve_fit


##########################################
def adaptive_screening(peaks_LF_all,peaks_HF_all_scr,peaks_HF_all_true,Tdata,acqtype,maxit,prange,sampinit,initplevels,dur,ne_exp,filefolder,figfolder,casename,varsymb,varunit,debugplotswitch):
    """

    Parameters
    ----------
    This function performs adaptive screening with both single-fidelity HF-GPR 
    and multi-fidelity GPR, for a case study of linear & second-order waves.
    Four cases are defined for this case study: perfect indicator, no noise, 
    LongTp noise and ShortTp noise. Refer to OMAE2023 and JOMAE2023/24 paper 
    for a description of the cases. Each can be evaluated with different adaptive
    sampling acquisition functions and max number of iterations. Outcome is a 
    plot of the predicted HF DNR compared to the true HF data and predictions 
    of the MPM and its uncertainty.
        
    peaks_LF_all : 0D array
        All LF indicator peaks in the Monte-Carlo simulation
    peaks_HF_all_scr : 0D array
        All HF peaks in the Monte-Carlo simulation that are matched to the LF 
        peaks, in exactly the same order. These are normally not available, as 
        they need to be derived using a CFD calculation or experiment for the 
        new event that is defined in each iteration.
    peaks_HF_all_true : 0D array
        All HF peaks in the validation Monte-Carlo simulation. These are usually 
        not available in a real strongly non-linear response application either.
        They are only used as ground truth for the validation.   
    Tdata : float
        Duration of the present dataset in seconds. Used together with ne_exp 
        to estimate the number of wave encounters in this duration (Nwaves). 
        This in turn is used as a basis for the probability of exceedance (PoE) 
        levels in all LF and HF distributions. Both the largest LF and HF value 
        have a PoE equal to 1/Nwaves.
    acqtype : string
        Acquisition function used for adaptive Generalised Pareto fitting
        - 'MV'      = Predicted value guidance (towards higher values)
        - 'MD'      = Function that simply selects the location within prange 
                      that is furthest from any of the earlier sample points.
    maxit : integer
        Max. number of iterations (on top of 3 initial samples)
    prange : 1D array
        Array providing the probability of exceedance prediction range. The GPR 
        or MF-GPR exceedance probability distribution will be predicted over 
        this range. Defined from low to high probability values. Example: 
        prange = np.linspace(plow,phigh,pstep)[:, None]
        The new samples will always be selected from within this range.
    sampinit : string
        Initial sample option for LF and HF data. Different options, that only 
        differ in their definition of the LF samples used by MF-GPR. The options
        are identical for GPR:
        - 'A' = initial HF samples defined at <initplevels> - LF sampling for 
                MF-GPR using resampling with one sample per 0.05 m interval
        - 'B' = initial HF samples defined at <initplevels> - LF sampling for 
                MF-GPR using GEV fit to LF samples over Prange_predLF
        - 'C' = initial HF samples defined at <initplevels> - LF sampling for 
                MF-GPR using GEV fit to available HF samples over Prange_predLF 
                (updated every iteration) - not tested yet.
    initplevels : 0D array
        Array of probability of exceedance levels at which we derive the initial 
        HF sampling. Typically has 2, 3, or 4 elements. These have to be within 
        Prange. Example: initplevels = np.array([p1,p2,p3]) in the case of three 
        initial samples. These probability levels may be outside prange (although 
        it may make sense to use e.g. the lower and upper limits of prange).
    dur : float
        Duration of interest / exposure duration. Typically between 30 min and 
        3 hours. Provided in seconds, so between 1800 and 10800.
    ne_exp : float
        Average number of wave encounters in the duration of interest <dur>. 
        Used to derive the probability of exceedance of interest and basis for 
        the PoE distribution axes.
    filefolder : string
        Folder name where to save the output CSV files (relative to present folder). 
        Location should exist already.
    figfolder : string
        Folder name where to save the output PDF figures (relative to present folder). 
        Location should exist already.
    casename : string
        Customisable name of the present case - will appear in the plot file names.
    varsymb : string
        Symbols for the LF and HF variable on the x-axis of the PoE plots. Have
        to be lower-case. Format: ['LF variable symbol','HF variable symbol']
    varunit : string
        Units for the variable on the x-axis of the probability of exceedance 
        plots. Have to be lower-case. Format: ['LF unit','HF unit']
    debugplotswitch : integer
        1 = plot (extra) debugging plots
        0 = do not plot these (extra) debugging plots
    
    Returns
    -------   
    itrangeGP : numpy array
        Number of samples included in each iteration of the GP fit procedure.
    ass_MPM_GP : numpy array
        MPM derived in each iteration of the GP fit procedure.
    v_pred_h : numpy array
        Value range for fitted results
    GP_exceedanceP_cor : numpy array
        Generalised Pareto fit to the HF samples in the last iteration, defined 
        over v_pred_h.
    
    
    Revision history
    ---------
    v1 Mar'24   Sanne van Essen     Connected to draft journal paper with a case 
                                    study of second-order waves, submitted to RESS
    v2 ??       Sanne van Essen     Connected to the updated journal paper.
    v3 Apr'24   Sanne van Essen     Version used for a new case study (SCREAM 
                                    experiments, predicting MPM values of non-linear
                                    RWE and / or wave impact loads)
    v3.1 Apr'24   Sanne van Essen   'Branch' version that uses Generalised Pareto
                                    fitting instead of (MF-)GPR.
    
    
    """
      
    # Figure sizes (see matplotlib doc)
    fsize_h = 7
    fsize_v = 5
    
    # Variable symbols and units
    varsymb_lf = varsymb[0]
    varsymb_hf = varsymb[1]
    varunit_lf = varunit[0]
    varunit_hf = varunit[1]
    
    # Check wether output directories exist, if not make them
    isExist = os.path.exists(filefolder)
    if not isExist:
       os.makedirs(filefolder)
    isExist = os.path.exists(figfolder)
    if not isExist:
       os.makedirs(figfolder)
    
    # Max value (for plotting) 
    minV = np.floor(0.9*np.nanmin(peaks_HF_all_scr)) # HF
    maxV = np.ceil(1.1*np.nanmax(peaks_HF_all_scr)) # HF
    maxV_lf = np.ceil(1.1*np.nanmax(peaks_LF_all)) # LF
          
    # Exposure duration probability level in the DNR
    pexp = 1/ne_exp  
    
    # Estimated number of wave encounters in full experimental duration (where 
    # the full experimental duration is estimated based on the time of the last 
    # indicator peak). Because the LF and HF signals can be different variables, 
    # we need a common basis for their probability of exceedance (PoE) distributions. 
    # This is provided by  Nwaves; the largest LF and HF value both have PoE 
    # 1/Nwaves in the output.
    Nwaves = Tdata/dur*ne_exp
            
    # # Hardcoded: Matern32 kernels for the MF-GPR computation    
    # kern_MFgp_lf = GPy.kern.Matern32(input_dim=1, variance=1., lengthscale=1.); MFkern1 = 'Mat32'
    # kern_MFgp_delta = GPy.kern.Matern32(input_dim=1, variance=1., lengthscale=1.); MFkern2 = 'Mat32'
    # kernels_MFgp = [kern_MFgp_lf,kern_MFgp_delta]
       
    # # Hardcoded: Matern32 kernel for the GPR computation    
    # kernel_HFgp = GPy.kern.Matern32(input_dim=1, variance=1., lengthscale=1.); HFkern = 'Mat32'
    
    # Prediction value range for the LF predictions in MF-GPR
    Prange_predLF = prange.copy()
    Prange_predLF_log = np.log(Prange_predLF)
    
    # Prediction value range for the HF predictions in GPR and MF-GPR
    Prange_predHF = prange.copy()
    Prange_predHF_log = np.log(Prange_predHF)
  
    # Instead, change the data points with zero HF response values from the 
    # matched data back to NaN values (these wave events) led to HF responses  
    # below the POT GP_threshold. Let's see how GPR accounts for this.
    i_zero_hf_scr = np.where(peaks_HF_all_scr==0)[0]
    peaks_HF_all_scr[i_zero_hf_scr] = np.nan
        
    ## ------- Derived distribution of true data
    
    # Calculate true distributions for 'Monte-Carlo' dataset (all data, validation) - also obtain 'screening' order indices
    v_true_l,p_true_l_temp,i_screening = calc_distribution(peaks_LF_all)
    v_true_h,p_true_h_temp,i_HFtrue = calc_distribution(peaks_HF_all_true) # Not available normally
     
    # Correct for basis Nwaves instead of the number of LF or HF peaks
    N_lf = len(peaks_LF_all)
    N_hf = len(peaks_HF_all_true)
    p_true_l = p_true_l_temp*N_lf/Nwaves
    p_true_h = p_true_h_temp*N_hf/Nwaves
    
    # Calculate true MPM values
    MPM_true_MCS = MPM_from_DNR(v_true_h,p_true_h,ne_exp)
       
    # Subsample the MCS true distribution data for plotting later
    binsizeHF = maxV/500
    binsHF = np.arange(0,maxV,binsizeHF)
    v_true_h_sampled,p_true_h_sampled = subsample_binned(v_true_h,p_true_h,binsHF)
    binsizeLF = maxV_lf/500
    binsLF = np.arange(0,maxV_lf,binsizeLF)
    v_true_l_sampled,p_true_l_sampled = subsample_binned(v_true_l,p_true_l,binsLF)
            
    # Max number of available LF datapoints (for DEM = Nseeds) 
    Nmax = len(v_true_l)

    # ------- Show difference true HF distribution and best possible screening result with all events and present indicator

    # # For reference; best distribution that could be achieved with present indicator and re-ordering
    # v_scr_h_reorder,p_scr_h_reorder = calc_distribution(peaks_HF_all_scr)
    
    # Screening result in the probability of exceedence distribution domain        
    v_scr_h = peaks_HF_all_scr[i_screening]
    p_scr_h = p_true_l.copy()
    
    # Find probability of exceedence level where half of the points is above and below (based on the LF indicator)
    i_half = np.int(np.ceil(len(v_true_l)/2))
    p_half = p_true_l[i_half]
    
    # Corresponding exposure duration (depends on the present LF MCS duration, 
    # response non-linearity and  maybe indicator quality)
    dur_half = pexp*dur/p_half
    
    # create bins and calculate bin mean
    nbins = 10
    Pbins_log = np.linspace(np.nanmin(np.log(p_scr_h)),np.nanmax(np.log(p_scr_h)),nbins)
    Pbins = np.exp(Pbins_log)
    Pbins_middle,v_scr_h_smooth4 = bin_means_smoothing(p_scr_h,v_scr_h,Pbins)
        
    plt.figure(6000,figsize=(fsize_h,fsize_v))
    plt.xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_hf,varunit_hf), fontsize=12)
    plt.ylabel(r'$\mathrm{P(}\mathrm{%s} \geq \mathrm{%s} \mathrm{)}$' % (varsymb_hf.upper(),varsymb_hf), fontsize=12)
    plt.xlabel(f'{varsymb_hf} [{varunit_hf}]')
    # plt.ylim([10**(-5),1])
    plt.xlim([0,maxV])
    plt.plot(v_scr_h,p_scr_h,marker='.',linestyle='none',color='purple',linewidth=0.8,label='All HF screening samples')
    plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',linestyle='--',color='pink',linewidth=0.8,label='HF "true" MCS')
    plt.plot(v_scr_h_smooth4,Pbins_middle,'-.',color='r',label=f'Smoothing by bin means ({nbins} bins)')
    plt.plot([0,maxV],[p_half,p_half],'--k',linewidth=0.5,label=r'$\mathrm{P}_{\mathrm{half}} \mathrm{ (}{%i}\mathrm{s)}$' % (np.int(dur_half)))
    plt.plot([0,maxV],[pexp,pexp],':',color='black',label=r'$\mathrm{P}_{\mathrm{exp}} \mathrm{ (}{%i}\mathrm{s)}$' % (np.int(dur)))
    plt.savefig(f'{figfolder}/ScreeningSamples_{casename}_ylin.pdf',bbox_inches='tight')  
    plt.yscale('log')
    plt.legend()
    plt.savefig(f'{figfolder}/ScreeningSamples_{casename}_ylog.pdf',bbox_inches='tight')  
      
    # ------- All LF samples (this leads to too many LF peaks for multiple realisations)
    v_true_l_t = v_true_l[~np.isnan(v_true_l)]
    p_true_l_t = p_true_l[~np.isnan(v_true_l)]
    v_scr_h_t = v_scr_h[~np.isnan(v_true_l)]
    p_scr_h_t = p_scr_h[~np.isnan(v_true_l)]

    
    if sampinit == 'A':
        # HF initial samples for GPR and MF-GPR defined at <initplevels>
        # LF samples for MF-GPR defined by subsampling all LF samples.
        txtadd = f'A_LFsubsamp_exp{np.int(dur/60)}min'
                   
        # Subsample the MCS true distribution data for plotting later
        binsize1 = maxV/100
        bins1 = np.arange(0,maxV,binsize1)
        v_train_l,p_train_l = subsample_binned(v_true_l_t,p_true_l_t,bins1)
           
        # Initialise 
        i_sel_dist = np.zeros(np.shape(initplevels),dtype=int)
        
        # Find corresponding indices of these samples (may be outside available pool)
        for cc,p_lev in enumerate(initplevels):
            i_sel_dist[cc],_ = find_nearest(p_true_l_t,p_lev)
           
        # Selected LF points where we pick the matched HF points (may be outside available pool)
        p_train_l_REF = p_true_l_t[i_sel_dist]
        v_train_l_REF = v_true_l_t[i_sel_dist]
        
        # Now we have to get back to the original time series peak matching to 
        # find the corresponding HF values > these are the HF training data values
        v_train_h = np.zeros(np.size(v_train_l_REF))
        for ii,selvalue in enumerate(v_train_l_REF):
            i_sel_time = np.where(peaks_LF_all == selvalue)
            v_train_h[ii] = peaks_HF_all_scr[i_sel_time]
            
        # HF probabilities are set identical to LF probabilities
        p_train_h = p_train_l_REF    
    
    
    elif sampinit == 'B':
        # HF initial samples for GPR and MF-GPR defined at <initplevels>
        # LF samples for MF-GPR defined by a Weibull fit to the topperc% highest LF datapoints.
        txtadd = f'B_LFfit_exp{np.int(dur/60)}min'

        # Generate 3-param Weibull fit to all LF data
        N_lf_points = 50 # number of LF points over the range indicated by topperc
        topperc = 50 # percentage of peaks to include in fit
        maxit_weib = 100 # maximum number of iterations
        _,_,theta,alpha,beta,corr = fit_weibull3_lsq(p_true_l_t,v_true_l_t,topperc,maxit_weib)
        vmin = v_true_l_t[-int(np.floor(len(v_true_l_t) * (topperc / 100)))]
        vmax = np.max(v_true_l_t) 
        v_topperc = np.linspace(vmin,vmax,N_lf_points)[:, None]
        p_weib_LF = np.exp(-((v_topperc - theta)/alpha)**beta)
                      
        # LF samples to use in prediction
        v_train_l = np.squeeze(v_topperc)
        p_train_l = np.squeeze(p_weib_LF)
        
        # Write LF fit to CSV file (required for fig 8)
        dat_wei = np.transpose(np.stack((np.squeeze(v_train_l),np.squeeze(p_train_l))))
        my_df_wei = pd.DataFrame(dat_wei)
        head_wei = ['Vweibull','Pweibull']
        my_df_wei.to_csv(f'{filefolder}/{casename}_LFsamples_3pWeibullfit_to_top{topperc}p.csv',header = head_wei, index= False)
               
        # Initialise 
        i_sel_dist = np.zeros(np.shape(initplevels),dtype=int)
        
        # Find corresponding indices of these samples (may be outside available pool)
        for cc,p_lev in enumerate(initplevels):
            i_sel_dist[cc],_ = find_nearest(p_true_l_t,p_lev)
            
        # Selected LF points where we pick the matched HF points (may be outside available pool)
        p_train_l_REF = p_true_l_t[i_sel_dist]
        v_train_l_REF = v_true_l_t[i_sel_dist]
        
        # Now we have to get back to the original time series peak matching to 
        # find the corresponding HF values > these are the HF training data values
        v_train_h = np.zeros(np.size(v_train_l_REF))
        for ii,selvalue in enumerate(v_train_l_REF):
            i_sel_time = np.where(peaks_LF_all == selvalue)
            v_train_h[ii] = peaks_HF_all_scr[i_sel_time]
            
        # HF probabilities are set identical to LF probabilities
        p_train_h = p_train_l_REF    
    
    elif sampinit == 'C':
        # HF initial samples for GPR and MF-GPR defined at <initplevels>
        # LF samples for MF-GPR defined by a Weibull fit to the available HF 
        # points. This fit needs to be updated every iteration, so the LF 
        # datapoints also change every iteration. This is handled later in the 
        # code. This option was not yet tested, so may have bugs.
        txtadd = f'C_LFisHFfit_exp{np.int(dur/60)}min'
        
        # Initialise 
        i_sel_dist = np.zeros(np.shape(initplevels),dtype=int)
        
        # Find corresponding indices of these samples (may be outside available pool)
        for cc,p_lev in enumerate(initplevels):
            i_sel_dist[cc],_ = find_nearest(p_true_l_t,p_lev)
           
        # Selected LF points where we pick the matched HF points (may be outside available pool)
        p_train_l_REF = p_true_l_t[i_sel_dist]
        v_train_l_REF = v_true_l_t[i_sel_dist]
        
        # Now we have to get back to the original time series peak matching to 
        # find the corresponding HF values > these are the HF training data values
        v_train_h = np.zeros(np.size(v_train_l_REF))
        for ii,selvalue in enumerate(v_train_l_REF):
            i_sel_time = np.where(peaks_LF_all == selvalue)
            v_train_h[ii] = peaks_HF_all_scr[i_sel_time]
            
        # HF probabilities are set identical to LF probabilities
        p_train_h = p_train_l_REF    
                
        # Generate 3-param Weibull fit to all available HF datapoints
        N_lf_points = 50 # number of LF points over the range indicated by topperc
        topperc = 100 # percentage of peaks to include in fit - all of available ones
        maxit_weib = 100 # maximum number of iterations
        _,_,theta,alpha,beta,corr = fit_weibull3_lsq(p_train_h,v_train_h,topperc,maxit_weib)
        vmin = v_train_h[-int(np.floor(len(v_true_l_t) * (topperc / 100)))]
        vmax = np.max(v_train_h) 
        v_topperc = np.linspace(vmin,vmax,N_lf_points)[:, None]
        p_weib_LF = np.exp(-((v_topperc - theta)/alpha)**beta)
       
        # LF samples to use in prediction
        v_train_l = np.squeeze(v_topperc)
        p_train_l = np.squeeze(p_weib_LF)

    # Make a pool of samples, within the prediction range prange (to avoid picking samples outside this range later)
    inds_range = np.where((p_true_l_t>=min(prange)) & (p_true_l_t<=max(prange)))[0]
    p_initsamplepool = p_true_l_t[inds_range]
    v_initsamplepool = v_true_l_t[inds_range]
    v_scrsamplepoolref = v_scr_h_t[inds_range]
    
    # Check and print how many events in the HF MCS are within prediction range
    # i_plow = np.where(p_true_h<=min(prange))[0][0]
    # i_phigh = np.where(p_true_h<=max(prange))[0][0]
    # noevents_in_range = i_plow-i_phigh
    noevents_in_range = len(inds_range)
    noevents_in_range_nonnan = len(v_scrsamplepoolref[~np.isnan(v_scrsamplepoolref)])
    print(f'There are {noevents_in_range} events within probability range "prange", of which {noevents_in_range_nonnan} HF events are not equal to NaN')
    if noevents_in_range_nonnan < (maxit+len(initplevels)):
        raise Exception(f'The number of non-NaN HF events within probability range "prange" ({noevents_in_range_nonnan}) is lower than the number of iterations specified (for either GPR or MF-GPR) plus the number of initial samples. In other words, there are not enough events within the specified range.') 
    
    
    # If initial samples are inside the available pool, exclude them to choose from later in adaptive sampling
    # Important: use prange limits here instead of sample pool limits, to avoid selecting the same point later (which in turn can introduce GPR issues)
    for cc,p_lev in enumerate(initplevels):
        # print(f'Element {cc}: p level {p_lev}')
        if (p_lev >= min(prange)) & (p_lev <= max(prange)):
            i_sel_dist_pool,_ = find_nearest(p_initsamplepool,p_lev)
            p_initsamplepool = np.delete(p_initsamplepool,i_sel_dist_pool)
            v_initsamplepool = np.delete(v_initsamplepool,i_sel_dist_pool)

    print(f'Initial sampling done: {txtadd}')
    
    ## ------- Plot initial true functions & training data
    # Separately for HF-GPR and MF-GPR, as different samples will be added in the
    # adaptive sampling procedures for both. Update these plots later with every 
    # iteration (add the new points).
    
    N_hf = len(v_train_h)
    
    # Subsample the true and screening data specifically for the scatter plot
    xbinsize = maxV/500
    ybinsize = maxV/500
    xbins = np.arange(0,maxV,xbinsize)
    ybins = np.arange(0,maxV,ybinsize)
    peaks_LF_all_sampled,peaks_HF_all_scr_sampled = subsample_binned_2D(peaks_LF_all,peaks_HF_all_scr,xbins,ybins)
    
    #  Scatter plot indicator vs NL response peaks, including initial samples
    plt.figure(50,figsize=(fsize_h,0.8*fsize_v))
    plt.yscale('linear')
    plt.plot(peaks_LF_all_sampled,peaks_HF_all_scr_sampled,marker='o',color='grey',label='All peaks',linestyle = 'None')
    # plt.plot(np.array([0,maxV_lf]),np.array([0,maxV]),'--',color='grey',label='ref')
    # plt.xlabel('LF indicator peaks (linear crests)')
    # plt.ylabel('Corr. HF response peaks (2nd-order crests)')
    plt.xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_lf.upper(),varunit_lf), fontsize=12)
    plt.ylabel(r'Corresponding $\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_hf.upper(),varunit_hf), fontsize=12)
    plt.xlim([0,maxV_lf]); plt.ylim([0,maxV])
    plt.plot(v_train_l_REF,v_train_h,marker='x',color='red',markersize = 10,label='HF samples',linestyle = 'None')
    for cc in range(N_hf):
        plt.text(v_train_l_REF[cc]+0.05,v_train_h[cc],'init')
    plt.legend()
    plt.savefig(f'{figfolder}/GP_Scatter_{casename}_{txtadd}_InitSamps.pdf',bbox_inches='tight')

    # Scale factor HF to LF
    v_train_h_MOD = v_train_h/maxV*maxV_lf
    
    # True and training data plot, including initial samples (use secondary x-axis for training data LF and HF)
    fig88 = plt.figure(880,figsize=(fsize_h,fsize_v))
    plt.yscale('log')
    plt.ylabel(r'$\mathrm{P(}\mathrm{%s} \geq \mathrm{%s} \mathrm{)}$   and   $\mathrm{P(}\mathrm{%s} \geq \mathrm{%s} \mathrm{)}$' % (varsymb_hf.upper(),varsymb_hf,varsymb_lf.upper(),varsymb_lf), fontsize=12)
    # -- x-axis1 (HF data)
    aX1a = plt.subplot()
    lnd1 = aX1a.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" MCS (normally N/A)')
    lnd2 = aX1a.plot(v_train_h,p_train_h,linestyle='None',color='r',marker='d',label='Resulting HF samples')
    lnd3 = aX1a.plot([0,maxV],[pexp,pexp],':',color='black') #,label='$P_{exp}$')
    aX1a.text(14,pexp*1.4,'$P_{exp}$')
    aX1a.plot(v_train_h,p_train_h,linestyle='None',color='r',marker='d') # plot repeated on top, no label
    aX1a.set_xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_hf,varunit_hf), fontsize=12, color='red')
    aX1a.set_xlim(left=0,right=maxV)
    # -- x-axis2 (LF data)
    aX2a = aX1a.twiny()
    lnd4 = aX2a.plot(v_true_l_sampled,p_true_l_sampled,marker='.',color=(0.7,0.7,0.7),linestyle='None',label='LF "true" MCS') 
    # lnd5 = aX2c.plot(v_train_l,p_train_l, color='grey', marker='o',linestyle='None', label='LF samples')
    ind6 = aX2a.plot(v_train_l_REF,p_train_l_REF,marker='x',markerfacecolor='none',color='grey',linestyle='None',label='LF basis for HF sampling')
    aX2a.set_xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_lf,varunit_lf), fontsize=12, color=(0.4,0.4,0.4))
    aX2a.set_xlim(left=0,right=maxV_lf)
    for bb,_ in enumerate(v_train_h): # add connection lines
        lnd7 = aX2a.plot([v_train_l_REF[bb],v_train_h_MOD[bb]],[p_train_l_REF[bb],p_train_h[bb]],linestyle='--',color='red',label='Screening connections')    
    for cc in range(N_hf):
        aX2a.text(v_train_l_REF[cc]+0.05,p_train_l_REF[cc],'init')  
    # -- y-axis
    # plt.ylim([10**(-5),1])
    # lnd = lnd4+lnd5+ind6+lnd7+lnd2+lnd1
    lnd = lnd4+ind6+lnd7+lnd2+lnd1
    labs = [l.get_label() for l in lnd]
    aX1a.legend(lnd,labs,loc='lower left')
    plt.savefig(f'{figfolder}/GenPar_TrainData_{casename}_{txtadd}_inclcon_InitSamps.pdf',bbox_inches='tight')    


    
    ## ------- Write true and initial training data to file
    
    # Prepare data
    dat1 = np.transpose(np.stack((np.squeeze(v_true_l),np.squeeze(p_true_l))))
    dat2 = np.transpose(np.stack((np.squeeze(v_true_h),np.squeeze(p_true_h))))
    dat3 = np.transpose(np.stack((np.squeeze(v_train_l),np.squeeze(p_train_l))))
    dat4 = np.transpose(np.stack((np.squeeze(v_train_h),np.squeeze(p_train_h))))
    
    # Write CSV file using pandas
    my_df1 = pd.DataFrame(dat1)
    my_df1.to_csv(f'{filefolder}/TrueLFdata_{casename}_{txtadd}.csv',header = False, index= False)
    my_df2 = pd.DataFrame(dat2)
    my_df2.to_csv(f'{filefolder}/TrueHFdata_{casename}_{txtadd}.csv',header = False, index= False)
    my_df3 = pd.DataFrame(dat3)
    my_df3.to_csv(f'{filefolder}/TrainLFdata_{casename}_{txtadd}.csv',header = False, index= False)
    my_df4 = pd.DataFrame(dat4)
    my_df4.to_csv(f'{filefolder}/TrainHFdata_init_{casename}_{txtadd}.csv',header = False, index= False)
    
    
    print('Writing data to files done')
    
    
    
    ## ------- Modify true functions and training data to emphasise tail (operate on log)
    
    p_true_l_log = np.log(p_true_l)
    p_true_h_log = np.log(p_true_h)
    p_train_l_log = np.log(p_train_l)
    p_train_h_log = np.log(p_train_h)
        
    # Adaptive sampling
    p_initsamplepool_log = np.log(p_initsamplepool)
    
    
    
    ## ------- 0D to vertical 1D arrays, as required by the GPy / Emukit toolboxes
    v_true_l = np.transpose([v_true_l])
    p_true_l = np.transpose([p_true_l])
    p_true_l_log = np.transpose([p_true_l_log])
    v_train_l = np.transpose([v_train_l])
    p_train_l = np.transpose([p_train_l])
    p_train_l_log = np.transpose([p_train_l_log])
    v_true_h = np.transpose([v_true_h])
    p_true_h = np.transpose([p_true_h])
    p_true_h_log = np.transpose([p_true_h_log])
    v_train_h = np.transpose([v_train_h])
    p_train_h = np.transpose([p_train_h])
    p_train_h_log = np.transpose([p_train_h_log])
    
    print('All preparations for (MF-)GPR done')
    print('From here we start with adaptive sampling, separately for HF-GPR and MF-GPR')
    
    
    #########################################
    ## ------- Iterations 
    
    if maxit != 0:
        
        # Prepare CSV file where MPM results will be stored over iterations
        csv_filename = f'{filefolder}/Conv_GenPar_{casename}_{txtadd}_{acqtype}_Nhf{N_hf}.csv'
        headGP = np.array(['NoSamp','MPM_GP','U95_MPM_GP','SumVar_GP','MeanVar_GP'])
        np.savetxt(csv_filename,[headGP],delimiter=',',fmt='%s')
        
        # Debug plot: check that sample pool is within prange
        if debugplotswitch == 1:
            plt.figure(100000000)
            plt.plot(v_initsamplepool,p_initsamplepool_log,'.-')
            plt.plot([0,14],[np.min(Prange_predHF_log),np.min(Prange_predHF_log)],'--k')
            plt.plot([0,14],[np.max(Prange_predHF_log),np.max(Prange_predHF_log)],'--k')
            plt.title('GPR: selected points on log(p)')
                
        # Initialise lists and arrays to save HF-GPR results as function of number of loops
        p_train_h_arr_log = []
        v_train_h_arr = []
        GP_arr_p = []
        GP_arr_v = []
        ass_MPM_GP = np.array([])
        ass_fit_R2 = np.array([])
        
        # Initialise
        count1 = 0
       
        # Start while loop
        while count1+N_hf <= Nmax and count1 <= maxit:
            
            ####  Fit Generalised Pareto distribution to the present (1-CDF) samples
            
            # Define value range over which the fitted function will be defined
            v_pred_h = np.linspace(minV,maxV,100)

            # Perform curve fitting - if GP fit does not converge, we use a linear fit instead (with a warning)
            try:
                
                # Define function
                function_gp = GP_pexc_log
                
                # Define data to fit the function to 
                xdata = np.squeeze(v_train_h)
                ydata = np.squeeze(p_train_h_log)
                                
                # Define parameter bounds (based on what we know about the distribution)
                # Order [loc, scale, shape], as in GP_pexp_log. Bounds not on log scale.
                lo_low = -np.inf
                # lo_high = -np.nanmax(v_train_h) # has to cross y-axis at value below this value (find out whether this bound can de set even lower) 
                lo_high = -np.nanmin(v_train_h) # has to cross y-axis at value below this value (find out whether this bound can de set even lower) 
                sc_low = -np.inf
                sc_high = np.inf
                sh_low = 0
                sh_high = 2 # could be set to shape parameter that does not deviate too much from 0 (forcing a near-exponential distribution), but not combined with location limits
                bounds_known=([lo_low, sc_low, sh_low], [lo_high, sc_high, sh_high])
                
                # Add hard constraint points at the lower and upper edges (if we can define them?)
                # ydatamod = 
                
                # Create the weighting array that forces the solution to be very close to the 'fixed' start and end points
                # (https://stackoverflow.com/questions/49151096/how-to-force-scipy-optimize-curve-fit-fix-first-point)
                y_weight = np.empty(len(ydata))
                #high pseudo-sd values, meaning less weighting in the fit
                y_weight.fill(10)
                #low values for point 0 and the last points, meaning more weighting during the fit procedure 
                y_weight[0] = y_weight[-1] = 0.1
                
                # Perform curve fitting
                # parameters, covariance_fit = curve_fit(function_gp, xdata, ydata)
                parameters, covariance_fit = curve_fit(function_gp, xdata, ydata, bounds=bounds_known)
                # parameters, covariance_fit = curve_fit(function_gp, xdata, ydata, bounds=bounds_known, sigma=y_weight) # Alternative with fixed points
                
                # Extract outcome
                loc = parameters[0]; scale = parameters[1];  shape = parameters[2]
                p_pred_h_log = function_gp(v_pred_h, loc, scale, shape)
                p_pred_h = np.exp(p_pred_h_log)
                
                # # Extract standard deviation errors on the parameters
                # err_fit_params = np.sqrt(np.diag(covariance_fit))
                
                # Calculate residuals (on linear scale, not log scale)
                residuals_fit = np.squeeze(p_train_h) - GP_pexc(xdata, *parameters)
                
                # Label plots 
                labplotfit = 'Gen. Pareto fit (lsq on log scale)'
                colplot = 'g'
                                
            except:
                
                # Warning
                print(f'WARNING: least-squares Generalised Pareto fitting to log(P) data does not converge for the present {len(v_train_h)} HF samples.')
                print(f'Instead, we use a linear least-squares fit on log scale (resulting in an exponential PoE distribution).')
                
                # # Perform linear fitting
                # DATX = np.vstack([np.squeeze(v_train_h), np.ones(len(v_train_h))]).T
                # a_lin,b_lin = np.linalg.lstsq(DATX,p_train_h_log,rcond=None)[0]
                # p_pred_h_log = a_lin*v_pred_h+b_lin
                # p_pred_h = np.exp(p_pred_h_log)
                
                # # Label plots 
                # labplotfit = 'Exponential fit (linear lsq fit on log scale)  - alternative to GP fit that did not converge'
                # colplot = 'c'
                
                # Perform 3-parameter Weibull fitting
                tp = 100 # percentage of peaks to include in fit (all of them)
                maxit_weib = 100 # maximum number of iterations
                _,_,theta_weib,alpha_weib,beta_weib,corr_weib = fit_weibull3_lsq(p_train_h,v_train_h,tp,maxit_weib)
                p_pred_h = np.exp(-((v_pred_h - theta_weib)/alpha_weib)**beta_weib)
                p_pred_h_log = np.log(p_pred_h)
                
                # Label plots 
                labplotfit = '3p Weibull fit (lsq) - alternative to GP fit that did not converge'
                
                # Calculate residuals (on linear scale, not log scale)
                residuals_fit = np.squeeze(p_train_h) - np.exp(-((v_train_h - theta_weib)/alpha_weib)**beta_weib)
                colplot = 'purple'
                
                
            # Calculate the R2 value using the residual sum of squares and total sum of squares
            # Residual sum of squares 
            ss_res = np.sum(residuals_fit**2)
            ss_tot = np.sum((ydata-np.nanmean(ydata))**2)
            R2 = 1 - (ss_res / ss_tot)
            
            # # Debugging of residual plots
            # plt.figure()
            # plt.plot(v_train_h,p_train_h,marker='d',linestyle='none',color='red',label=f'HF samples iteration {count1}')
            # plt.plot(v_pred_h,p_pred_h,label=f'Generalised Pareto fit (least-squares on log scale)')
            # plt.title(f'{count1}')
            
            # Derive MPM from this fit (uncertainty not available)
            MPM_GP = np.interp(pexp,np.squeeze(p_pred_h[::-1]),np.squeeze(v_pred_h[::-1]))    
            
            
            # Debug plots: fitting of the GP distribution to the empirical 1-CFD
            if debugplotswitch == 1:
                
                # Plot distribution
                plt.figure()
                plt.yscale('log')
                # plt.plot(v_train_h, np.exp(p_train_h_log), 'o', label='data') # should match
                plt.plot(v_train_h,p_train_h,marker='d',linestyle='none',color='red',label=f'HF samples iteration {count1}')
                plt.plot(v_pred_h,p_pred_h,label=labplotfit)
                plt.plot([0,maxV],[pexp,pexp],':',color='black',label='$P_{exp}$')
                plt.plot([MPM_GP,MPM_GP],[10**(-5),1],'--k',label='MPM')
                plt.title(f'Iteration {count1}')
                plt.xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_hf,varunit_hf), fontsize=12)
                plt.ylabel(r'$\mathrm{P(}\mathrm{%s} \geq \mathrm{%s} \mathrm{)}$' % (varsymb_hf.upper(),varsymb_hf), fontsize=12)
                # plt.ylim([10**(-5),1])
                plt.xlim([0,maxV])
                plt.legend()
                
            # Save full results of present iteration to list
            p_train_h_arr_log.append(p_train_h_log)
            v_train_h_arr.append(v_train_h)
            GP_arr_v.append(v_pred_h)
            GP_arr_p.append(p_pred_h)
    
            # Save MPM and goodness of fit metric to array
            ass_MPM_GP = np.append(ass_MPM_GP,MPM_GP)
            ass_fit_R2 = np.append(ass_fit_R2,R2)
            
            ## ------- Write present training data and prediction to file

            # Prepare data
            p_train_h = np.exp(p_train_h_log)
            dat_trainHF = np.transpose(np.stack((np.squeeze(v_train_h),np.squeeze(p_train_h))))
            dat_predHF_GP  = np.transpose(np.stack((np.squeeze(v_pred_h),np.squeeze(p_pred_h))))
        
            # Write CSV file for present iteration using pandas
            df_trainHF = pd.DataFrame(dat_trainHF)
            head_trainHF_hfpgr = ['v_train_h','p_train_h']
            df_trainHF.to_csv(f'{filefolder}/GenPar_TrainData_{casename}_{txtadd}_acq-{acqtype}_Nhf{N_hf}_it{count1}.csv',header = head_trainHF_hfpgr, index= False)
            
            df_predHF_GP = pd.DataFrame(dat_predHF_GP)
            head_predHF_GP = ['V_genpareto','Pexc']
            df_predHF_GP.to_csv(f'{filefolder}/GenPar_HFpred_{casename}_{txtadd}_acq-{acqtype}_Nhf{N_hf}_it{count1}.csv',header = head_predHF_GP, index= False)
                
            # Append convergence data to file (iteration number, MPM, U95 of MPM, sumvar and meanvar)
            nosamps = N_hf+len(ass_MPM_GP)-1
            newdat_conv_GP = np.array([nosamps,MPM_GP])
            with open(csv_filename,'a') as file:
                np.savetxt(file, [newdat_conv_GP], delimiter=',', fmt='%s')

            # Do this for each loop, except the last one (we don't need a new sample in that case)
            # Defining it this way saves the double notation of the whole GPR procedure (before & inside the loop)
            # (Only works for the maximum # iterations stopping criterion)
            if count1 < maxit:
                               
                # Define acquisition function Adaptive Screening with Generalised Pareto
                if acqtype == 'MD':
                    func_acquisition = MaxDist(p_train_h_log)
                # elif acqtype == 'MV':
                     # ADDD 
                                
                # Now evaluate this acquisition function on the target probability range
                func_acq_plot = func_acquisition.evaluate(Prange_predHF_log)
                
                # Debugging
                if debugplotswitch == 1:
                    plt.figure()
                    plt.plot(Prange_predHF_log,func_acq_plot)
                    plt.title(f'GPR, {count1}')
                
                # Based on results, define a new sample
                # To find the next point to evaluate we optimize the acquisition function using 
                # a standard gradient descent optimizer. It seems like this finds the coordinate 
                # of the max. value of the acq function in 1D.
                # optimizer = GradientAcquisitionOptimizer(EMU_space_log) # SOMETIMES GOES WRONG
                # p_new_sample_log_target, _ = optimizer.optimize(func_acquisition) # SOMETIMES GOES WRONG
                p_new_sample_log_target = Prange_predHF_log[np.argmax(func_acq_plot),0]
               
                # Find new point that is not equal to NaN 
                v_train_h_new = np.nan
                while np.isnan(v_train_h_new): 
                    # Find closest sample in the LF screening MCS result
                    i_new_sample_log,p_new_sample_log = find_nearest(p_initsamplepool_log,p_new_sample_log_target)
                    # print(i_new_sample_log)
                    
                    # Selected LF points where we pick the matched HF points 
                    p_train_l1_log_REF_new = p_initsamplepool_log[i_new_sample_log] # = p_new_sample_log
                    v_train_l1_REF_new = v_initsamplepool[i_new_sample_log]
                    
                    # Now we have to get back to the original time series peak matching to find the
                    # corresponding HF value > this is the HF sample
                    i_sel_time_new = np.where(peaks_LF_all == v_train_l1_REF_new)
                    v_train_h_new = peaks_HF_all_scr[i_sel_time_new]
                        
                    # HF probabilities are set identical to LF probabilities
                    p_train_h_log_new = p_train_l1_log_REF_new.copy()
                    
                    # Remove newly chosen sample from the pool of samples for the next iteration
                    p_initsamplepool_log = np.delete(p_initsamplepool_log,i_new_sample_log)
                    v_initsamplepool = np.delete(v_initsamplepool,i_new_sample_log) 
                
                # Debug
                if debugplotswitch == 1:
                    print(f'Iteration {count1}: new point close to log(p) = {p_new_sample_log_target}')
                    plt.figure(100000000)
                    plt.plot(v_train_l1_REF_new,p_train_l1_log_REF_new,'xr')
                    plt.text(v_train_l1_REF_new,p_train_l1_log_REF_new,f'{count1}') #,color='r')
                            
                print(f'Adaptive Screening with Generalised Pareto fitting: new sample number {count1} is defined')
            
                # Add the new samples to the scatter plot
                plt.figure(50)
                plt.plot(v_train_l1_REF_new,v_train_h_new,marker='x',color='red',markersize = 10,linestyle = 'None')
                plt.text(v_train_l1_REF_new+0.05,v_train_h_new,f'New sample {count1+1}')
                            
                # Scale factor HF to LF
                v_train_h_new_MOD = v_train_h_new/maxV*maxV_lf
                
                # Add the new samples to the training data plot
                plt.figure(880)
                aX2a.plot(v_train_l1_REF_new,np.exp(p_train_l1_log_REF_new),marker='x',markerfacecolor='none',color='grey',linestyle='None') # ,label='LF basis for HF sampling')
                aX1a.text(v_train_h_new+0.3,np.exp(p_train_h_log_new),f'it{count1+1}',color='r')
                aX2a.plot([v_train_l1_REF_new,v_train_h_new_MOD],np.exp([p_train_l1_log_REF_new,p_train_h_log_new]),linestyle='--',color='red')
                aX1a.plot(v_train_h_new,np.exp(p_train_h_log_new),linestyle='None',color='r',marker='d') # ,label='Resulting HF samples')
                
                # Append new sample to existing HF samples
                p_train_h_log = np.append(p_train_h_log,p_train_h_log_new)
                v_train_h = np.append(v_train_h,v_train_h_new)
                
                # Add dimension such that we have a column array
                p_train_h_log = p_train_h_log[:,None]
                v_train_h = v_train_h[:,None]
                
                # Make sure that the non-log version keeps matching
                p_train_h = np.exp(p_train_h_log)
                            
            # --------- Add to counter to continue the while loop, unless we reached the max. # seeds
            count1 = count1+1
            # if count1+N_hf > Nmax:
            #     print("Warning: loop stops because max. # of available seeds is reached in HF-GPR adaptive sampling loop, while uncertainties are still above tolerance.")
                
        # Number of last iteration
        lastcount1 = count1-1
        
        # Inverse the log of the final samples and results
        p_train_h = np.exp(p_train_h_log)
                
        # Save final scatter plot and sample distribution plot
        if lastcount1 > 5:
            plt.figure(50)
            plt.savefig(f'{figfolder}/GenPar_Scatter_{casename}_{txtadd}_{acqtype}_Nhf{N_hf}_it{lastcount1}.pdf',bbox_inches='tight')
            
            plt.figure(880)
            plt.savefig(f'{figfolder}/GenPar_TrainData_{casename}_{txtadd}_inclcon_{acqtype}_Nhf{N_hf}_it{lastcount1}.pdf',bbox_inches='tight')
        
        ###### For reference: 3-param Weibull fit to the final HF samples
        topperc = 100 # percentage of peaks to include in fit (all of them)
        maxit_weib_fin = 100 # maximum number of iterations
        _,_,theta_fin,alpha_fin,beta_fin,corr_fin = fit_weibull3_lsq(p_train_h,v_train_h,topperc,maxit_weib_fin)
        p_weib_HF_fin = np.exp(-((v_pred_h - theta_fin)/alpha_fin)**beta_fin)
        
        ###### For reference: linear fit to the final HF samples
        A1 = np.vstack([np.squeeze(v_train_h), np.ones(len(v_train_h))]).T
        m1,linfitstart1 = np.linalg.lstsq(A1,p_train_h,rcond=None)[0]
        p_lin_ref1 = m1*v_pred_h+linfitstart1
        
        ###### For reference: linear fit to the final HF samples on LOG scale (moved exponential distribution)
        A2 = np.vstack([np.squeeze(v_train_h), np.ones(len(v_train_h))]).T
        m2,linfitstart2 = np.linalg.lstsq(A2,p_train_h_log,rcond=None)[0]
        p_lin_ref2_log = m2*v_pred_h+linfitstart2
        p_lin_ref2 = np.exp(p_lin_ref2_log)
        
        
        # Plot the data and final predictions - original data
        fig200 = plt.figure(2000,figsize=(fsize_h,fsize_v))
        plt.yscale('log')
        plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None') #,label='HF "true" MCS')
        plt.plot(v_train_h,p_train_h, color='r', marker='d',linestyle='None',markersize=4, label='HF samples')
        plt.plot(v_pred_h,p_pred_h,'--',color=colplot,label=labplotfit)
        if labplotfit.find('Weibull') == -1:
            plt.plot(v_pred_h,p_weib_HF_fin,'-.',color='purple',label='For ref.: 3-parameter Weibull fit to final HF samples (lsq)')
        # plt.plot(v_pred_h,p_lin_ref1,':y',label='For reference: linear fit to final HF samples')
        if labplotfit.find('Linear') == -1:
            plt.plot(v_pred_h,p_lin_ref2,':c',label='For ref.: exponential fit to final HF samples (linear lsq fit on log scale)')
        plt.plot([0,maxV],[pexp,pexp],':',color='black',linewidth=0.5) #,label='$P_{exp}$')
        plt.plot(v_true_h_sampled[0],p_true_h_sampled[0],marker='.',color='pink',linestyle='None',label='HF "true" MCS (validation)') # To put this one at the end of the legend
        plt.text(14,pexp*1.4,'$P_{exp}$')
        plt.xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_hf,varunit_hf), fontsize=12)
        plt.ylabel(r'$\mathrm{P(}\mathrm{%s} \geq \mathrm{%s} \mathrm{)}$ w.r.t. # wave encounters' % (varsymb_hf.upper(),varsymb_hf), fontsize=12)
        # plt.ylim([10**(-5),1])
        # plt.ylim([10**(-4),10**(-1)])
        plt.xlim([0,maxV])
        plt.legend()
        if lastcount1 > 5:
            plt.savefig(f'{figfolder}/Prediction_GenPar_{casename}_{txtadd}_{acqtype}_Nhf{N_hf}_it{lastcount1}.pdf',bbox_inches='tight')
        
        
        
        
        ###### For reference: Generalised Pareto fit to the re-ordered HF samples
        doreorder = 0 # This is not a good idea. Maybe if we select very close to Pexp and make a DEM. Or start selection based on DEM. 
        
        if doreorder == 1:
        
            # Re-order final HF samples
            v_train_h_reord,p_train_h_reord_t,i_reord = calc_distribution(np.squeeze(v_train_h))
            
            # Corresponding log version
            p_train_h_log_reord_t = np.log(p_train_h_reord_t)
            
            # Fit GP function directly to the log of the PoE distribution
            function_gp = GP_pexc_log
            xdata_r = np.squeeze(v_train_h_reord)
            ydata_r = np.squeeze(p_train_h_log_reord_t)
            parameters_r, covariance_fit_r = curve_fit(function_gp, xdata_r, ydata_r)
            loc_r = parameters_r[0]; scale_r = parameters_r[1];  shape_r = parameters_r[2]
            p_pred_h_log_reord_t = function_gp(v_pred_h, loc_r, scale_r, shape_r)
            p_pred_h_reord_t = np.exp(p_pred_h_log_reord_t)
    
            # Correct for PoE axis (keep original bounds)
            Prefsamparray = [1 / i for i in range(1,len(v_train_h_reord)+1)] # in 'our' system
            Prefnew = np.nanmax(Prefsamparray)
            Preforig = np.nanmax(p_train_h)
            ratio1 = Preforig/Prefnew
            p_train_h_reord = p_train_h_reord_t*ratio1     
            p_pred_h_reord = p_pred_h_reord_t*ratio1    
                   
            # # Fit quality
            # # err_fit_params = np.sqrt(np.diag(covariance_fit_r))
            # residuals_fit_r = np.exp(ydata_r) - GP_pexc(xdata_r, *parameters_r)
            # ss_res_r = np.sum(residuals_fit_r**2)
            # ss_tot_r = np.sum((ydata_r-np.nanmean(ydata_r))**2)
            # R2_r = 1 - (ss_res_r / ss_tot_r)
            
            # Plot the data and final predictions - re-ordered final samples
            fig300 = plt.figure(3000,figsize=(fsize_h,fsize_v))
            plt.yscale('log')
            plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None') #,label='HF "true" MCS')
            plt.plot(v_train_h,p_train_h, color='r', marker='d',linestyle='None',markersize=4,label='Screening HF samples') # alpha=.5
            if 'Pareto' in labplotfit:
                plt.plot(v_pred_h,p_pred_h,'--',color=colplot,label='Gen. Pareto fit to screening samples (lsq on log scale)')
            elif 'Weibull' in labplotfit:
                plt.plot(v_pred_h,p_pred_h,'--',color=colplot,label='3p Weibull fit to screening samples (lsq)')
            elif 'Linear' in labplotfit:
                plt.plot(v_pred_h,p_pred_h,'--',color=colplot,label='Exponential fit to final HF samples (linear lsq fit on log scale)')  
            plt.plot(v_train_h_reord,p_train_h_reord, color='darkturquoise', marker='o',linestyle='None',markersize=4, label='Re-ordered HF samples')
            plt.plot(v_pred_h,p_pred_h_reord,'--',color='darkturquoise',label='Gen. Pareto fit to re-ordered samples (lsq on log scale)')
            plt.plot([0,maxV],[pexp,pexp],':',color='black',linewidth=0.5) #,label='$P_{exp}$')
            plt.plot(v_true_h_sampled[0],p_true_h_sampled[0],marker='.',color='pink',linestyle='None',label='HF "true" MCS (validation)') # To put this one at the end of the legend
            plt.text(14,pexp*1.4,'$P_{exp}$')
            plt.xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_hf,varunit_hf), fontsize=12)
            plt.ylabel(r'$\mathrm{P(}\mathrm{%s} \geq \mathrm{%s} \mathrm{)}$ w.r.t. # wave encounters' % (varsymb_hf.upper(),varsymb_hf), fontsize=12)
            # plt.ylim([10**(-5),1])
            plt.ylim([10**(-4),10**(-1)])
            plt.xlim([0,maxV])
            plt.legend()
            if lastcount1 > 5:
                plt.savefig(f'{figfolder}/Prediction_GenPar_ReOrdSamp_{casename}_{txtadd}_{acqtype}_Nhf{N_hf}_it{lastcount1}.pdf',bbox_inches='tight')
                 
        
        
        print('Adaptive Screening with Generalised Pareto fitting (MLE fit) done')
        
        ## -------  Visualise iterations and convergence
        
        # Visualise iterations and convergence
        itrangeGP = np.arange(N_hf,N_hf+len(ass_MPM_GP),1)
        seedmaxplot = maxit+N_hf
        
        plt.figure(figsize=(6,4))
        plt.plot(itrangeGP,ass_MPM_GP,'-',color='b',label=f'Adaptive Screening / Gen. Pareto fit (acq. {acqtype})')
        plt.plot([0,seedmaxplot],[MPM_true_MCS,MPM_true_MCS],':r',label='True HF MPM (from MCS)')
        plt.ylabel('MF MPM [m]')
        plt.xlim([N_hf-1,seedmaxplot])
        plt.xlabel('Number of HF samples')
        plt.xticks(np.arange(N_hf-1,seedmaxplot,1,dtype=int))
        plt.tight_layout()
        plt.legend()
        if lastcount1 > 5:
            plt.savefig(f'{figfolder}/Convergence_ASwithGP_{casename}_{txtadd}_{acqtype}_Nhf{N_hf}_it{lastcount1}.pdf',bbox_inches='tight')       
            
            
            
            
        print('Everything done')
        
    
    return itrangeGP,ass_MPM_GP,v_pred_h,p_pred_h



##########################################   
def present_hyperGP_params(gpr_model):
    """
    Access the names and values of the present hyperparameters of a GPR object
    gpr_model (made with toolbox GPy) before or after optimization, and print 
    their names and values.
    """
    # Assuming high_gp_model is your GPR model
    values = gpr_model.param_array
    
    # Get the names of the hyperparameters
    names = gpr_model.parameter_names()
    
    # # Print names and values of the optimized hyperparameters
    # for name1, value1 in zip(names, values):
    #     print(f"{name1}: {value1}")
        
    return names,values   
    

##########################################
def find_nearest(array, value):
       
    array = np.asarray(array)
    # idx = (np.abs(array - value)).argmin()
    temp = (np.abs(array - value))
    i_nearest = np.nanargmin(temp)
    
    nearest = array[i_nearest]
        
    return i_nearest,nearest


##########################################
def bin_means_smoothing(x,y,Xbins):
    # Initialize an array to store smoothed values
    smoothed_y = np.zeros(len(Xbins)-1)
    Xbins_middle = np.zeros(len(Xbins)-1)
    
    # Iterate over bins and calculate the mean value of y within each bin
    for i in range(len(Xbins) - 1):
        mask = np.logical_and(x >= Xbins[i], x < Xbins[i+1])
        mean_y = np.nanmean(y[mask])
        smoothed_y[i] = mean_y
        Xbins_middle[i] = (Xbins[i] + Xbins[i+1])/2
    
    return Xbins_middle,smoothed_y


# ##########################################
# Function to fit Generalized Pareto Distribution (GPD) to empirical CDF
def fit_gpd_to_cdf(values, probabilities, GP_threshold):
    # Compute excesses
    excesses = values[values > GP_threshold] - GP_threshold

    # Estimate parameters using Maximum Likelihood Estimation (MLE)
    GP_params = genpareto.fit(excesses)

    return GP_params


# ##########################################
# Function to fit Generalized Pareto Distribution (GPD) to peak over GP_threshold (POT) observations
def fit_gpd_to_pot(observations, GP_threshold):
    # Extract POT observations
    pot_observations = observations[observations > GP_threshold]

    # Fit GPD parameters using Maximum Likelihood Estimation (MLE)
    GP_params = genpareto.fit(pot_observations - GP_threshold)

    return GP_params


# ##########################################
# Empirical CDF (newly ordered, used to debug the GP fit
def empirical_cdf(observations, GP_threshold):
    # Compute empirical CDF
    pot_observations = observations[observations >= GP_threshold]
    pot_observations = np.sort(pot_observations)
    ecdf_values = np.arange(1, len(pot_observations) + 1) / len(pot_observations)
    return pot_observations, ecdf_values



